# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=protected-access

import errno
import unittest
from io import StringIO
from unittest.mock import MagicMock, call, patch

from pontos.testing import temp_directory
from rich.console import Console

from greenbone.feed.sync.errors import FileLockingError, GreenboneFeedSyncError
from greenbone.feed.sync.helper import (
    Spinner,
    change_user_and_group,
    flock_wait,
    is_root,
)


class FlockTestCase(unittest.IsolatedAsyncioTestCase):
    async def test_locking(self):
        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"
            self.assertFalse(lock_file.exists())

            async with flock_wait(lock_file):
                self.assertTrue(lock_file.exists())

                with self.assertRaises(FileLockingError):
                    async with flock_wait(lock_file, wait_interval=None):
                        pass

    async def test_can_not_create_parent_dirs(self):
        with temp_directory() as temp_dir:
            lock_file = temp_dir / "foo" / "file.lock"
            temp_dir.chmod(0)

            with self.assertRaisesRegex(
                FileLockingError, "Could not create parent directories for "
            ):
                async with flock_wait(lock_file):
                    pass

    async def test_console_output(self):
        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"
            console = MagicMock(spec=Console)

            async with flock_wait(lock_file, console=console):
                pass

            console.print.assert_has_calls(
                [
                    call(f"Trying to acquire lock on {lock_file.absolute()}"),
                    call(f"Acquired lock on {lock_file.absolute()}"),
                    call(f"Releasing lock on {lock_file.absolute()}"),
                ]
            )

    @patch("greenbone.feed.sync.helper.fcntl.flock", autospec=True)
    async def test_retry(self, flock_mock: MagicMock):
        e = OSError()
        e.errno = errno.EACCES

        flock_mock.side_effect = [e, None, None]

        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"
            console = MagicMock(spec=Console)

            async with flock_wait(
                lock_file,
                console=console,
                wait_interval=0.5,
            ):
                pass

            console.print.assert_has_calls(
                [
                    call(f"Trying to acquire lock on {lock_file.absolute()}"),
                    call(
                        f"{lock_file.absolute()} is locked by another process."
                        " Waiting 0.5 seconds before next try."
                    ),
                    call(f"Trying to acquire lock on {lock_file.absolute()}"),
                    call(f"Acquired lock on {lock_file.absolute()}"),
                    call(f"Releasing lock on {lock_file.absolute()}"),
                ]
            )

    @patch("greenbone.feed.sync.helper.fcntl.flock", autospec=True)
    async def test_lock_failure(self, flock_mock: MagicMock):
        e = OSError()
        e.errno = errno.EACCES

        flock_mock.side_effect = e

        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"

            with self.assertRaisesRegex(
                FileLockingError,
                f"{lock_file.absolute()} is locked. Another process related "
                "to the feed update may already running.",
            ):
                async with flock_wait(
                    lock_file,
                    wait_interval=None,
                ):
                    pass

    @patch("greenbone.feed.sync.helper.fcntl.flock", autospec=True)
    async def test_lock_other_failure(self, flock_mock: MagicMock):
        e = OSError("Other OSError")
        flock_mock.side_effect = e

        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"

            with self.assertRaisesRegex(OSError, "Other OSError"):
                async with flock_wait(
                    lock_file,
                    wait_interval=None,
                ):
                    pass

    @patch("greenbone.feed.sync.helper.fcntl.flock", autospec=True)
    async def test_unlock_oserror(self, flock_mock: MagicMock):
        flock_mock.side_effect = [None, OSError]

        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"
            async with flock_wait(
                lock_file,
            ):
                pass

    async def test_permission_error(self):
        with temp_directory() as temp_dir:
            lock_file = temp_dir / "file.lock"
            lock_file.touch()
            lock_file.chmod(0)
            with self.assertRaisesRegex(
                FileLockingError,
                f"^Permission error while trying to open the lock file {lock_file.absolute()}$",
            ):
                async with flock_wait(
                    lock_file,
                ):
                    pass


class SpinnerTestCase(unittest.TestCase):
    @patch("greenbone.feed.sync.helper.Live", autospec=True)
    def test_context_manager(self, live_mock: MagicMock):
        console = MagicMock(spec=Console)
        with Spinner(console, "Some Text"):
            pass

        live_mock_instance = live_mock.return_value
        live_mock_instance.start.assert_called_once_with()
        live_mock_instance.stop.assert_called_once_with()

    def test_render(self):
        out = StringIO()
        console = Console(file=out)
        with Spinner(console, "Some Text"):
            pass

        self.assertEqual("⠋ Some Text", out.getvalue())


class IsRootTestCase(unittest.TestCase):
    @patch("greenbone.feed.sync.helper.os.geteuid", autospec=True)
    def test_not_root(self, geteuid_mock: MagicMock):
        geteuid_mock.return_value = 123

        self.assertFalse(is_root())

    @patch("greenbone.feed.sync.helper.os.geteuid", autospec=True)
    def test_root(self, geteuid_mock: MagicMock):
        geteuid_mock.return_value = 0

        self.assertTrue(is_root())


class ChangeUserAndGroupTestCase(unittest.TestCase):
    @patch("greenbone.feed.sync.helper.os", autospec=True)
    def test_change(self, os_mock: MagicMock):
        # root user should exist on all systems
        change_user_and_group("root", "root")

        os_mock.seteuid.assert_called_once_with(0)
        os_mock.setegid.assert_called_once_with(0)

    @patch("greenbone.feed.sync.helper.shutil", autospec=True)
    @patch("greenbone.feed.sync.helper.os", autospec=True)
    def test_change_with_unknown_user(
        self, os_mock: MagicMock, shutil_mock: MagicMock
    ):
        shutil_mock._get_uid.return_value = None

        with self.assertRaisesRegex(
            GreenboneFeedSyncError, "Can't run as user 'foo'.*"
        ):
            change_user_and_group("foo", "bar")

        os_mock.seteuid.assert_not_called()
        os_mock.setegid.assert_not_called()

    @patch("greenbone.feed.sync.helper.shutil", autospec=True)
    @patch("greenbone.feed.sync.helper.os", autospec=True)
    def test_change_with_unknown_group(
        self, os_mock: MagicMock, shutil_mock: MagicMock
    ):
        shutil_mock._get_uid.return_value = 123
        shutil_mock._get_gid.return_value = None

        with self.assertRaisesRegex(
            GreenboneFeedSyncError, "Can't run as group 'bar'.*"
        ):
            change_user_and_group("foo", "bar")

        os_mock.seteuid.assert_not_called()
        os_mock.setegid.assert_not_called()



# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Optional, Sequence

import shtab

from greenbone.feed.sync.__version__ import __version__
from greenbone.feed.sync.config import (
    DEFAULT_CONFIG_FILE,
    DEFAULT_USER_CONFIG_FILE,
    Config,
    ConfigDict,
    EnterpriseSettings,
    maybe_int,
)
from greenbone.feed.sync.errors import ConfigFileError


def _to_defaults(values: ConfigDict) -> dict[str, Any]:
    defaults = {}

    for key, value in values.items():
        defaults[key.replace("-", "_")] = value

    return defaults


def feed_type(value: str) -> str:
    """
    Converts to a specific feed type
    """

    value = value.replace("_", "-").lower()

    if value in ("nvts", "report-formats", "port-lists", "scan-configs"):
        return value[:-1]

    return value


class CliParser:
    """
    An ArgumentParser for the feed sync CLI
    """

    def __init__(self) -> None:
        parser = ArgumentParser(add_help=False)
        shtab.add_argument_to(parser)
        parser.add_argument(
            "--version",
            help="Print version then exit.",
            action="version",
            version=f"%(prog)s {__version__}",
        )
        parser.add_argument(
            "-h",
            "--help",
            help="Show this help message and exit.",
            action="store_true",
        )
        parser.add_argument(
            "--selftest",
            help="Perform self-test and set exit code",
            action="store_true",
        )
        output_group = parser.add_mutually_exclusive_group()
        output_group.add_argument(
            "--verbose",
            "-v",
            action="count",
            help="Set log verbosity. `-vvv` for maximum verbosity. "
            "(Default: %(default)s)",
        )
        output_group.add_argument(
            "--quiet", action="store_true", help="Disable all log output."
        )

        parser.add_argument(
            "-c",
            "--config",
            help="Configuration file path. If not set %(prog)s "
            f"tries to load {DEFAULT_USER_CONFIG_FILE} or "
            f"{DEFAULT_CONFIG_FILE}.",
        )
        parser.add_argument(
            "--private-directory",
            help="(Sub-)Directory to exclude from the sync which will never "
            "get deleted automatically.",
            type=Path,
        )
        parser.add_argument(
            "--compression-level",
            type=int,
            choices=range(0, 10),
            help="Rsync compression level (0-9). (Default: %(default)s)",
        )
        parser.add_argument(
            "--type",
            choices=[
                "all",
                "nvt",
                "gvmd-data",
                "scap",
                "cert",
                "notus",
                "nasl",
                "report-format",
                "scan-config",
                "port-list",
            ],
            default="all",
            type=feed_type,
            help="Select which feed should be synced. (Default: %(default)s)",
        )
        parser.add_argument(
            "--feed-version",
            help="Deprecated. Use --feed-release instead.",
        )
        parser.add_argument(
            "--feed-release",
            help="Release series of the feed to download for example 22.04 or 24.10. (Default: %(default)s)",
        )
        parser.add_argument(
            "--destination-prefix",
            type=Path,
            help="Prefix for the destination directories. (Default: %(default)s)",
        )
        parser.add_argument(
            "--gvmd-data-destination",
            type=Path,
            help="Destination of the downloaded gvmd data. "
            "(Default: %(default)s)",
        )
        parser.add_argument(
            "--gvmd-data-url",
            help="URL to download the gvmd data from. (Default: %(default)s)",
        )
        nvts_destination_group = parser.add_argument_group()
        nvts_destination_group.add_argument(
            "--notus-destination",
            type=Path,
            help="Destination of the downloaded notus data. "
            "(Default: %(default)s)",
        )
        nvts_url_group = parser.add_argument_group()
        nvts_url_group.add_argument(
            "--notus-url",
            help="URL to download the notus data from. (Default: %(default)s)",
        )
        nvts_destination_group.add_argument(
            "--nasl-destination",
            type=Path,
            help="Destination of the downloaded nasl data. "
            "(Default: %(default)s)",
        )
        nvts_url_group.add_argument(
            "--nasl-url",
            help="URL to download the nasl data from. (Default: %(default)s)",
        )

        secinfo_destination_group = parser.add_argument_group()
        secinfo_destination_group.add_argument(
            "--scap-data-destination",
            type=Path,
            help="Destination of the downloaded SCAP data. "
            "(Default: %(default)s)",
        )
        secinfo_url_group = parser.add_argument_group()
        secinfo_url_group.add_argument(
            "--scap-data-url",
            help="URL to download the SCAP data from. (Default: %(default)s)",
        )
        secinfo_destination_group.add_argument(
            "--cert-data-destination",
            type=Path,
            help="Destination of the downloaded CERT data. "
            "(Default: %(default)s)",
        )
        secinfo_url_group.add_argument(
            "--cert-data-url",
            help="URL to download the CERT data from. (Default: %(default)s)",
        )

        data_objects_destination_group = parser.add_argument_group()
        data_objects_destination_group.add_argument(
            "--report-formats-destination",
            type=Path,
            help="Destination of the downloaded report format data "
            "(Default: %(default)s)",
        )
        data_objects_url_group = parser.add_argument_group()
        data_objects_url_group.add_argument(
            "--report-formats-url",
            help="URL to download the report format data from. "
            "(Default: %(default)s)",
        )
        data_objects_destination_group.add_argument(
            "--scan-configs-destination",
            type=Path,
            help="Destination of the downloaded scan config data. "
            "(Default: %(default)s)",
        )
        data_objects_url_group.add_argument(
            "--scan-configs-url",
            help="URL to download the scan config data from. "
            "(Default: %(default)s)",
        )
        data_objects_destination_group.add_argument(
            "--port-lists-destination",
            type=Path,
            help="Destination of the downloaded port list data. "
            "(Default: %(default)s)",
        )
        data_objects_url_group.add_argument(
            "--port-lists-url",
            help="URL to download the port list data from. "
            "(Default: %(default)s)",
        )

        lock_file_group = parser.add_argument_group()
        lock_file_group.add_argument(
            "--gvmd-lock-file",
            type=Path,
            help="File to use for locking the feed synchronization for data "
            "loaded by the gvmd daemon. Used to avoid that more then one "
            "process accesses the feed data at the same time. "
            "(Default: %(default)s)",
        )
        lock_file_group.add_argument(
            "--openvas-lock-file",
            type=Path,
            help="File to use for locking the feed synchronization for data "
            "loaded by the openvas scanner. Used to avoid that more then one "
            "process accesses the feed data at the same time. "
            "(Default: %(default)s)",
        )
        parser.add_argument(
            "--fail-fast",
            "--failfast",
            action="store_true",
            help="Stop after a first error has occurred. Otherwise the script "
            "tries to download additional data if specified.",
        )

        wait_group = parser.add_mutually_exclusive_group()
        wait_group.add_argument(
            "--no-wait",
            action="store_true",
            help="Fail directly if the lock file can't be acquired.",
        )
        wait_group.add_argument(
            "--wait-interval",
            type=int,
            help="Time to wait in seconds after failed lock attempt before"
            "re-trying to lock the file. (Default: %(default)s seconds)",
        )

        parser.add_argument(
            "--rsync-timeout",
            type=int,
            help="Maximum I/O timeout in seconds used for rsync. If no data is "
            "transferred for the specified time then rsync will exit. By "
            "default no timeout is set and the rsync default will be used.",
        )

        permissions_group = parser.add_argument_group()
        permissions_group.add_argument(
            "--user",
            type=maybe_int,
            help="If started as root, use this user name or ID to run the "
            "script. (Default: %(default)s)",
        )
        permissions_group.add_argument(
            "--group",
            type=maybe_int,
            help="If started as root, use this group name or ID to run the "
            "script. (Default: %(default)s)",
        )

        parser.add_argument(
            "--greenbone-enterprise-feed-key",
            type=Path,
            help="File to read the Greenbone Enterprise Feed key from. "
            "The key gives access to additional vulnerability tests for "
            "enterprise software among other advantages. See "
            "https://www.greenbone.net/en/feed-comparison/ for more details."
            "The default URLs are adjusted according to the data in the key."
            "If the key file does not exist it is ignored. "
            "(Default: %(default)s)",
        )

        self.parser = parser

    def _determine_enterprise_settings(
        self, enterprise_key: Path
    ) -> Optional[EnterpriseSettings]:
        if not enterprise_key or not enterprise_key.exists():
            return None

        return EnterpriseSettings.from_key(enterprise_key)

    def _load_config(self, config_file: str) -> Config:
        config_path = None

        if config_file is None:
            for file in [DEFAULT_USER_CONFIG_FILE, DEFAULT_CONFIG_FILE]:
                path = Path(file).expanduser().resolve()
                if path.exists():
                    config_path = path
                    break

        else:
            config_path = Path(config_file).expanduser().resolve()
            if not config_path.exists():
                raise ConfigFileError(
                    f"Config file {config_file} does not exist."
                )

        config = Config()

        if config_path:
            config.load_from_config_file(config_path)

        return config

    def _set_defaults(self, config: ConfigDict) -> None:
        self.parser.set_defaults(**_to_defaults(config))

    def parse_arguments(
        self, args: Optional[Sequence[str]] = None
    ) -> Namespace:
        """
        Parse CLI arguments
        """
        # Parse args to get the config file path passed as option
        known_args, _ = self.parser.parse_known_args(args)

        # Load the defaults from the config file if it exists.
        config = self._load_config(known_args.config)

        # set greenbone enterprise feed key in config if user passed one to load
        # desired key for determining the feed url
        if known_args.greenbone_enterprise_feed_key:
            config["greenbone-enterprise-feed-key"] = (
                known_args.greenbone_enterprise_feed_key
            )

        if known_args.feed_version:
            config["feed-release"] = known_args.feed_version

        if known_args.feed_release:
            config["feed-release"] = known_args.feed_release

        if known_args.destination_prefix:
            config["destination-prefix"] = known_args.destination_prefix

        if self.parser.prog == "greenbone-nvt-sync":
            config["type"] = "nvt"
        elif self.parser.prog == "greenbone-scapdata-sync":
            config["type"] = "scap"
        elif self.parser.prog == "greenbone-certdata-sync":
            config["type"] = "cert"

        # apply defaults in config
        config.apply_settings()

        # check if a enterprise feed key is available
        enterprise_settings = self._determine_enterprise_settings(
            config["greenbone-enterprise-feed-key"]
        )

        # override feed url from key
        if enterprise_settings:
            config["feed-url"] = enterprise_settings.feed_url()

        # apply other config defaults
        config.apply_dependent_settings()

        # set config as defaults for CLI to make them visible in --help
        self._set_defaults(config)

        if known_args.help:
            self.parser.print_help()
            self.parser.exit(0)

        return self.parser.parse_args(args)



# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=line-too-long
# ruff: noqa: E501

import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.testing import temp_file

from greenbone.feed.sync.config import (
    DEFAULT_DESTINATION_PREFIX,
    DEFAULT_ENTERPRISE_KEY_PATH,
    DEFAULT_FEED_RELEASE,
    DEFAULT_GROUP,
    DEFAULT_GVMD_LOCK_FILE_PATH,
    DEFAULT_OPENVAS_LOCK_FILE_PATH,
    DEFAULT_USER,
    Config,
    EnterpriseSettings,
)
from greenbone.feed.sync.errors import ConfigError, ConfigFileError
from greenbone.feed.sync.helper import DEFAULT_FLOCK_WAIT_INTERVAL
from greenbone.feed.sync.rsync import (
    DEFAULT_RSYNC_COMPRESSION_LEVEL,
    DEFAULT_RSYNC_URL,
)


class ConfigTestCase(unittest.TestCase):
    def test_defaults(self):
        values = Config.load()

        self.assertEqual(len(values), 31)
        self.assertEqual(
            values["destination-prefix"], Path(DEFAULT_DESTINATION_PREFIX)
        )
        self.assertEqual(
            values["gvmd-data-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "data-objects" / "gvmd",
        )
        self.assertEqual(
            values["gvmd-data-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/",
        )
        self.assertEqual(values["feed-url"], DEFAULT_RSYNC_URL)
        self.assertEqual(
            values["notus-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "notus",
        )
        self.assertEqual(
            values["notus-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
        )
        self.assertEqual(
            values["nasl-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "openvas" / "plugins",
        )
        self.assertEqual(
            values["nasl-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
        )
        self.assertEqual(
            values["scap-data-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "scap-data",
        )
        self.assertEqual(
            values["scap-data-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/scap-data/",
        )
        self.assertEqual(
            values["cert-data-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "cert-data",
        )
        self.assertEqual(
            values["cert-data-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/cert-data/",
        )
        self.assertEqual(
            values["report-formats-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "report-formats",
        )
        self.assertEqual(
            values["report-formats-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/report-formats/",
        )
        self.assertEqual(
            values["scan-configs-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "scan-configs",
        )
        self.assertEqual(
            values["scan-configs-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/scan-configs/",
        )
        self.assertEqual(
            values["port-lists-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "port-lists",
        )
        self.assertEqual(
            values["port-lists-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/port-lists/",
        )
        self.assertEqual(
            values["gvmd-lock-file"],
            Path(DEFAULT_DESTINATION_PREFIX) / DEFAULT_GVMD_LOCK_FILE_PATH,
        )
        self.assertEqual(
            values["openvas-lock-file"],
            Path(DEFAULT_DESTINATION_PREFIX) / DEFAULT_OPENVAS_LOCK_FILE_PATH,
        )
        self.assertEqual(values["wait-interval"], DEFAULT_FLOCK_WAIT_INTERVAL)
        self.assertFalse(values["no-wait"])
        self.assertEqual(
            values["compression-level"], DEFAULT_RSYNC_COMPRESSION_LEVEL
        )
        self.assertIsNone(values["private-directory"])
        self.assertIsNone(values["verbose"])
        self.assertFalse(values["fail-fast"])
        self.assertIsNone(values["rsync-timeout"])
        self.assertEqual(values["group"], DEFAULT_GROUP)
        self.assertEqual(values["user"], DEFAULT_USER)
        self.assertEqual(
            values["greenbone-enterprise-feed-key"],
            Path(DEFAULT_ENTERPRISE_KEY_PATH),
        )
        self.assertEqual(values["feed-release"], DEFAULT_FEED_RELEASE)

    def test_config_file(self):
        content = """[greenbone-feed-sync]
destination-prefix = "/opt/lib"
feed-url = "rsync://lorem.ipsum"
gvmd-data-destination = "/usr/lib/gvmd-data"
gvmd-data-url = "rsync://foo.bar/gvmd-data"
notus-destination = "/usr/lib/notus"
notus-url = "rsync://foo.bar/notus"
nasl-destination = "/usr/lib/openvas/plugins/"
nasl-url = "rsync://foo.bar/nasl"
scap-data-destination = "/usr/lib/scap-data"
scap-data-url = "rsync://foo.bar/scap-data"
cert-data-destination = "/usr/lib/cert-data"
cert-data-url = "rsync://foo.bar/cert-data"
report-formats-destination = "/usr/lib/report-formats"
report-formats-url = "rsync://foo.bar/report-formats"
scan-configs-destination = "/usr/lib/scan-configs"
scan-configs-url = "rsync://foo.bar/scan-configs"
port-lists-destination = "/usr/lib/port-lists"
port-lists-url = "rsync://foo.bar/port-lists"
openvas-lock-file = "/usr/lib/openvas.lock"
gvmd-lock-file = "/usr/lib/gvmd.lock"
wait-interval = 100
no-wait = true
compression-level = 1
private-directory = "keep-this"
verbose = 5
fail-fast = true
rsync-timeout = 120
group = "foo"
user = "bar"
greenbone-enterprise-feed-key = "/srv/feed.key"
feed-release = "1.2.3"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        values = Config.load(path_mock)

        self.assertEqual(values["destination-prefix"], Path("/opt/lib"))
        self.assertEqual(values["feed-url"], "rsync://lorem.ipsum")
        self.assertEqual(
            values["gvmd-data-destination"], Path("/usr/lib/gvmd-data")
        )
        self.assertEqual(values["gvmd-data-url"], "rsync://foo.bar/gvmd-data")
        self.assertEqual(
            values["nasl-destination"], Path("/usr/lib/openvas/plugins")
        )
        self.assertEqual(values["notus-destination"], Path("/usr/lib/notus"))
        self.assertEqual(values["notus-url"], "rsync://foo.bar/notus")
        self.assertEqual(
            values["nasl-destination"], Path("/usr/lib/openvas/plugins")
        )
        self.assertEqual(values["nasl-url"], "rsync://foo.bar/nasl")
        self.assertEqual(
            values["scap-data-destination"], Path("/usr/lib/scap-data")
        )
        self.assertEqual(values["scap-data-url"], "rsync://foo.bar/scap-data")
        self.assertEqual(
            values["cert-data-destination"], Path("/usr/lib/cert-data")
        )
        self.assertEqual(values["cert-data-url"], "rsync://foo.bar/cert-data")
        self.assertEqual(
            values["report-formats-destination"],
            Path("/usr/lib/report-formats"),
        )
        self.assertEqual(
            values["report-formats-url"], "rsync://foo.bar/report-formats"
        )
        self.assertEqual(
            values["scan-configs-destination"],
            Path("/usr/lib/scan-configs"),
        )
        self.assertEqual(
            values["scan-configs-url"], "rsync://foo.bar/scan-configs"
        )
        self.assertEqual(
            values["port-lists-destination"],
            Path("/usr/lib/port-lists"),
        )
        self.assertEqual(values["port-lists-url"], "rsync://foo.bar/port-lists")
        self.assertEqual(
            values["openvas-lock-file"],
            Path("/usr/lib/openvas.lock"),
        )
        self.assertEqual(values["gvmd-lock-file"], Path("/usr/lib/gvmd.lock"))
        self.assertEqual(values["wait-interval"], 100)
        self.assertTrue(values["no-wait"])
        self.assertEqual(values["compression-level"], 1)
        self.assertEqual(values["private-directory"], Path("keep-this"))
        self.assertEqual(values["verbose"], 5)
        self.assertTrue(values["fail-fast"])
        self.assertEqual(values["rsync-timeout"], 120)
        self.assertEqual(values["group"], "foo")
        self.assertEqual(values["user"], "bar")
        self.assertEqual(
            values["greenbone-enterprise-feed-key"], Path("/srv/feed.key")
        )
        self.assertEqual(values["feed-release"], "1.2.3")

    def test_destination_prefix(self):
        content = """[greenbone-feed-sync]
destination-prefix = "/opt/lib/"
"""

        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        values = Config.load(path_mock)

        self.assertEqual(values["destination-prefix"], Path("/opt/lib"))
        self.assertEqual(
            values["gvmd-data-destination"],
            Path("/opt/lib/gvm/data-objects/gvmd/"),
        )
        self.assertEqual(values["notus-destination"], Path("/opt/lib/notus"))
        self.assertEqual(
            values["nasl-destination"], Path("/opt/lib/openvas/plugins")
        )
        self.assertEqual(
            values["scap-data-destination"], Path("/opt/lib/gvm/scap-data")
        )
        self.assertEqual(
            values["cert-data-destination"], Path("/opt/lib/gvm/cert-data")
        )
        self.assertEqual(
            values["report-formats-destination"],
            Path("/opt/lib/gvm/data-objects/gvmd/report-formats"),
        )
        self.assertEqual(
            values["scan-configs-destination"],
            Path("/opt/lib/gvm/data-objects/gvmd/scan-configs"),
        )
        self.assertEqual(
            values["port-lists-destination"],
            Path("/opt/lib/gvm/data-objects/gvmd/port-lists"),
        )
        self.assertEqual(
            values["openvas-lock-file"],
            Path("/opt/lib/openvas/feed-update.lock"),
        )
        self.assertEqual(
            values["gvmd-lock-file"], Path("/opt/lib/gvm/feed-update.lock")
        )

    def test_feed_url(self):
        content = """[greenbone-feed-sync]
feed-url = "rsync://foo.bar"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        values = Config.load(path_mock)

        self.assertEqual(values["feed-url"], "rsync://foo.bar")
        self.assertEqual(
            values["gvmd-data-url"],
            f"rsync://foo.bar/data-feed/{DEFAULT_FEED_RELEASE}/",
        )
        self.assertEqual(
            values["notus-url"],
            f"rsync://foo.bar/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
        )
        self.assertEqual(
            values["nasl-url"],
            f"rsync://foo.bar/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
        )
        self.assertEqual(
            values["scap-data-url"],
            f"rsync://foo.bar/vulnerability-feed/{DEFAULT_FEED_RELEASE}/scap-data/",
        )
        self.assertEqual(
            values["cert-data-url"],
            f"rsync://foo.bar/vulnerability-feed/{DEFAULT_FEED_RELEASE}/cert-data/",
        )
        self.assertEqual(
            values["report-formats-url"],
            f"rsync://foo.bar/data-feed/{DEFAULT_FEED_RELEASE}/report-formats/",
        )
        self.assertEqual(
            values["scan-configs-url"],
            f"rsync://foo.bar/data-feed/{DEFAULT_FEED_RELEASE}/scan-configs/",
        )
        self.assertEqual(
            values["port-lists-url"],
            f"rsync://foo.bar/data-feed/{DEFAULT_FEED_RELEASE}/port-lists/",
        )

    @patch.dict(
        "os.environ",
        {
            "GREENBONE_FEED_SYNC_DESTINATION_PREFIX": "/opt/lib",
            "GREENBONE_FEED_SYNC_URL": "rsync://lorem.ipsum",
            "GREENBONE_FEED_SYNC_GVMD_DATA_DESTINATION": "/usr/lib/gvmd-data",
            "GREENBONE_FEED_SYNC_GVMD_DATA_URL": "rsync://foo.bar/gvmd-data",
            "GREENBONE_FEED_SYNC_NOTUS_DESTINATION": "/usr/lib/notus",
            "GREENBONE_FEED_SYNC_NOTUS_URL": "rsync://foo.bar/notus",
            "GREENBONE_FEED_SYNC_NASL_DESTINATION": "/usr/lib/openvas/plugins/",
            "GREENBONE_FEED_SYNC_NASL_URL": "rsync://foo.bar/nasl",
            "GREENBONE_FEED_SYNC_SCAP_DATA_DESTINATION": "/usr/lib/scap-data",
            "GREENBONE_FEED_SYNC_SCAP_DATA_URL": "rsync://foo.bar/scap-data",
            "GREENBONE_FEED_SYNC_CERT_DATA_DESTINATION": "/usr/lib/cert-data",
            "GREENBONE_FEED_SYNC_CERT_DATA_URL": "rsync://foo.bar/cert-data",
            "GREENBONE_FEED_SYNC_REPORT_FORMATS_DESTINATION": "/usr/lib/report-formats",
            "GREENBONE_FEED_SYNC_REPORT_FORMATS_URL": "rsync://foo.bar/report-formats",
            "GREENBONE_FEED_SYNC_SCAN_CONFIGS_DESTINATION": "/usr/lib/scan-configs",
            "GREENBONE_FEED_SYNC_SCAN_CONFIGS_URL": "rsync://foo.bar/scan-configs",
            "GREENBONE_FEED_SYNC_PORT_LISTS_DESTINATION": "/usr/lib/port-lists",
            "GREENBONE_FEED_SYNC_PORT_LISTS_URL": "rsync://foo.bar/port-lists",
            "GREENBONE_FEED_SYNC_OPENVAS_LOCK_FILE": "/usr/lib/openvas.lock",
            "GREENBONE_FEED_SYNC_GVMD_LOCK_FILE": "/usr/lib/gvmd.lock",
            "GREENBONE_FEED_SYNC_LOCK_WAIT_INTERVAL": "100",
            "GREENBONE_FEED_SYNC_NO_WAIT": "1",
            "GREENBONE_FEED_SYNC_COMPRESSION_LEVEL": "1",
            "GREENBONE_FEED_SYNC_PRIVATE_DIRECTORY": "keep-this",
            "GREENBONE_FEED_SYNC_VERBOSE": "5",
            "GREENBONE_FEED_SYNC_FAIL_FAST": "1",
            "GREENBONE_FEED_SYNC_RSYNC_TIMEOUT": "120",
            "GREENBONE_FEED_SYNC_GROUP": "123",
            "GREENBONE_FEED_SYNC_USER": "321",
            "GREENBONE_FEED_SYNC_ENTERPRISE_FEED_KEY": "/tmp/some.key",
            "GREENBONE_FEED_SYNC_FEED_RELEASE": "1.2.3",
        },
    )
    def test_environment(self):
        values = Config.load()

        self.assertEqual(values["destination-prefix"], Path("/opt/lib"))
        self.assertEqual(values["feed-url"], "rsync://lorem.ipsum")
        self.assertEqual(
            values["gvmd-data-destination"], Path("/usr/lib/gvmd-data")
        )
        self.assertEqual(values["gvmd-data-url"], "rsync://foo.bar/gvmd-data")
        self.assertEqual(values["notus-destination"], Path("/usr/lib/notus"))
        self.assertEqual(values["notus-url"], "rsync://foo.bar/notus")
        self.assertEqual(
            values["nasl-destination"], Path("/usr/lib/openvas/plugins")
        )
        self.assertEqual(values["nasl-url"], "rsync://foo.bar/nasl")
        self.assertEqual(
            values["scap-data-destination"], Path("/usr/lib/scap-data")
        )
        self.assertEqual(values["scap-data-url"], "rsync://foo.bar/scap-data")
        self.assertEqual(
            values["cert-data-destination"], Path("/usr/lib/cert-data")
        )
        self.assertEqual(values["cert-data-url"], "rsync://foo.bar/cert-data")
        self.assertEqual(
            values["report-formats-destination"],
            Path("/usr/lib/report-formats"),
        )
        self.assertEqual(
            values["report-formats-url"], "rsync://foo.bar/report-formats"
        )
        self.assertEqual(
            values["scan-configs-destination"],
            Path("/usr/lib/scan-configs"),
        )
        self.assertEqual(
            values["scan-configs-url"], "rsync://foo.bar/scan-configs"
        )
        self.assertEqual(
            values["port-lists-destination"],
            Path("/usr/lib/port-lists"),
        )
        self.assertEqual(values["port-lists-url"], "rsync://foo.bar/port-lists")
        self.assertEqual(
            values["openvas-lock-file"],
            Path("/usr/lib/openvas.lock"),
        )
        self.assertEqual(values["gvmd-lock-file"], Path("/usr/lib/gvmd.lock"))
        self.assertEqual(values["wait-interval"], 100)
        self.assertTrue(values["no-wait"])
        self.assertEqual(values["compression-level"], 1)
        self.assertEqual(values["private-directory"], Path("keep-this"))
        self.assertEqual(values["verbose"], 5)
        self.assertTrue(values["fail-fast"])
        self.assertEqual(values["rsync-timeout"], 120)
        self.assertEqual(values["group"], 123)
        self.assertEqual(values["user"], 321)
        self.assertEqual(
            values["greenbone-enterprise-feed-key"], Path("/tmp/some.key")
        )
        self.assertEqual(values["feed-release"], "1.2.3")

    @patch.dict(
        "os.environ",
        {
            "GREENBONE_FEED_SYNC_DESTINATION_PREFIX": "/opt/lib",
            "GREENBONE_FEED_SYNC_URL": "rsync://lorem.ipsum",
            "GREENBONE_FEED_SYNC_GVMD_DATA_DESTINATION": "/usr/lib/gvmd-data",
            "GREENBONE_FEED_SYNC_GVMD_DATA_URL": "rsync://foo.bar/gvmd-data",
            "GREENBONE_FEED_SYNC_NOTUS_DESTINATION": "/usr/lib/notus",
            "GREENBONE_FEED_SYNC_NOTUS_URL": "rsync://foo.bar/notus",
            "GREENBONE_FEED_SYNC_NASL_DESTINATION": "/usr/lib/openvas/plugins/",
            "GREENBONE_FEED_SYNC_NASL_URL": "rsync://foo.bar/nasl",
            "GREENBONE_FEED_SYNC_SCAP_DATA_DESTINATION": "/usr/lib/scap-data",
            "GREENBONE_FEED_SYNC_SCAP_DATA_URL": "rsync://foo.bar/scap-data",
            "GREENBONE_FEED_SYNC_CERT_DATA_DESTINATION": "/usr/lib/cert-data",
            "GREENBONE_FEED_SYNC_CERT_DATA_URL": "rsync://foo.bar/cert-data",
            "GREENBONE_FEED_SYNC_REPORT_FORMATS_DESTINATION": "/usr/lib/report-formats",
            "GREENBONE_FEED_SYNC_REPORT_FORMATS_URL": "rsync://foo.bar/report-formats",
            "GREENBONE_FEED_SYNC_SCAN_CONFIGS_DESTINATION": "/usr/lib/scan-configs",
            "GREENBONE_FEED_SYNC_SCAN_CONFIGS_URL": "rsync://foo.bar/scan-configs",
            "GREENBONE_FEED_SYNC_PORT_LISTS_DESTINATION": "/usr/lib/port-lists",
            "GREENBONE_FEED_SYNC_PORT_LISTS_URL": "rsync://foo.bar/port-lists",
            "GREENBONE_FEED_SYNC_OPENVAS_LOCK_FILE": "/usr/lib/openvas.lock",
            "GREENBONE_FEED_SYNC_GVMD_LOCK_FILE": "/usr/lib/gvmd.lock",
            "GREENBONE_FEED_SYNC_LOCK_WAIT_INTERVAL": "100",
            "GREENBONE_FEED_SYNC_NO_WAIT": "1",
            "GREENBONE_FEED_SYNC_COMPRESSION_LEVEL": "1",
            "GREENBONE_FEED_SYNC_PRIVATE_DIRECTORY": "keep-this",
            "GREENBONE_FEED_SYNC_VERBOSE": "5",
            "GREENBONE_FEED_SYNC_FAIL_FAST": "1",
            "GREENBONE_FEED_SYNC_RSYNC_TIMEOUT": "120",
            "GREENBONE_FEED_SYNC_ENTERPRISE_FEED_KEY": "/tmp/some.key",
            "GREENBONE_FEED_SYNC_FEED_RELEASE": "1.2.3",
        },
    )
    def test_environment_overrides_config_file(self):
        content = """[greenbone-feed-sync]
destination-prefix = "/svr/lib"
feed-url = "rsync://ipsum.lorem"
gvmd-data-destination = "/root/notus"
gvmd-data-url = "rsync://bar.foo/notus"
notus-destination = "/root/notus"
notus-url = "rsync://bar.foo/notus"
nasl-destination = "/root/openvas/plugins/"
nasl-url = "rsync://bar.foo/nasl"
scap-data-destination = "/root/scap-data"
scap-data-url = "rsync://bar.foo/scap-data"
cert-data-destination = "/root/cert-data"
cert-data-url = "rsync://bar.foo/cert-data"
report-formats-destination = "/root/report-formats"
report-formats-url = "rsync://bar.foo/report-formats"
scan-configs-destination = "/root/scan-configs"
scan-configs-url = "rsync://bar.foo/scan-configs"
port-lists-destination = "/root/port-lists"
port-lists-url = "rsync://bar.foo/port-lists"
openvas-lock-file = "/root/openvas.lock"
gvmd-lock-file = "/root/gvmd.lock"
wait-interval = 99
no-wait = false
compression-level = 7
private-directory = "private"
verbose = 99
fail-fast = false
rsync-timeout = 360
greenbone-enterprise-feed-key = "/srv/feed.key"
feed-release = "2.3.4"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        values = Config.load(path_mock)

        self.assertEqual(values["destination-prefix"], Path("/opt/lib"))
        self.assertEqual(values["feed-url"], "rsync://lorem.ipsum")
        self.assertEqual(
            values["gvmd-data-destination"], Path("/usr/lib/gvmd-data")
        )
        self.assertEqual(values["gvmd-data-url"], "rsync://foo.bar/gvmd-data")
        self.assertEqual(values["notus-destination"], Path("/usr/lib/notus"))
        self.assertEqual(values["notus-url"], "rsync://foo.bar/notus")
        self.assertEqual(
            values["nasl-destination"], Path("/usr/lib/openvas/plugins")
        )
        self.assertEqual(values["nasl-url"], "rsync://foo.bar/nasl")
        self.assertEqual(
            values["scap-data-destination"], Path("/usr/lib/scap-data")
        )
        self.assertEqual(values["scap-data-url"], "rsync://foo.bar/scap-data")
        self.assertEqual(
            values["cert-data-destination"], Path("/usr/lib/cert-data")
        )
        self.assertEqual(values["cert-data-url"], "rsync://foo.bar/cert-data")
        self.assertEqual(
            values["report-formats-destination"],
            Path("/usr/lib/report-formats"),
        )
        self.assertEqual(
            values["report-formats-url"], "rsync://foo.bar/report-formats"
        )
        self.assertEqual(
            values["scan-configs-destination"],
            Path("/usr/lib/scan-configs"),
        )
        self.assertEqual(
            values["scan-configs-url"], "rsync://foo.bar/scan-configs"
        )
        self.assertEqual(
            values["port-lists-destination"],
            Path("/usr/lib/port-lists"),
        )
        self.assertEqual(values["port-lists-url"], "rsync://foo.bar/port-lists")
        self.assertEqual(
            values["openvas-lock-file"],
            Path("/usr/lib/openvas.lock"),
        )
        self.assertEqual(values["gvmd-lock-file"], Path("/usr/lib/gvmd.lock"))
        self.assertEqual(values["wait-interval"], 100)
        self.assertTrue(values["no-wait"])
        self.assertEqual(values["compression-level"], 1)
        self.assertEqual(values["private-directory"], Path("keep-this"))
        self.assertEqual(values["verbose"], 5)
        self.assertTrue(values["fail-fast"])
        self.assertEqual(values["rsync-timeout"], 120)
        self.assertEqual(
            values["greenbone-enterprise-feed-key"], Path("/tmp/some.key")
        )
        self.assertEqual(values["feed-release"], "1.2.3")

    def test_feed_release(self):
        feed_release = "1.2.3"
        content = f"""[greenbone-feed-sync]
feed-release = "{feed_release}"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        values = Config.load(path_mock)

        self.assertEqual(
            values["gvmd-data-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / feed_release,
        )
        self.assertEqual(
            values["gvmd-data-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{feed_release}/",
        )
        self.assertEqual(values["feed-url"], DEFAULT_RSYNC_URL)
        self.assertEqual(
            values["notus-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "notus",
        )
        self.assertEqual(
            values["notus-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{feed_release}/vt-data/notus/",
        )
        self.assertEqual(
            values["nasl-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "openvas" / "plugins",
        )
        self.assertEqual(
            values["nasl-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{feed_release}/vt-data/nasl/",
        )
        self.assertEqual(
            values["scap-data-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "scap-data",
        )
        self.assertEqual(
            values["scap-data-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{feed_release}/scap-data/",
        )
        self.assertEqual(
            values["cert-data-destination"],
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "cert-data",
        )
        self.assertEqual(
            values["cert-data-url"],
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{feed_release}/cert-data/",
        )
        self.assertEqual(
            values["report-formats-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / feed_release
            / "report-formats",
        )
        self.assertEqual(
            values["report-formats-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{feed_release}/report-formats/",
        )
        self.assertEqual(
            values["scan-configs-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / feed_release
            / "scan-configs",
        )
        self.assertEqual(
            values["scan-configs-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{feed_release}/scan-configs/",
        )
        self.assertEqual(
            values["port-lists-destination"],
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / feed_release
            / "port-lists",
        )
        self.assertEqual(
            values["port-lists-url"],
            f"{DEFAULT_RSYNC_URL}/data-feed/{feed_release}/port-lists/",
        )

    def test_invalid_feed_release(self):
        feed_release = "abc"
        content = f"""[greenbone-feed-sync]
feed-release = "{feed_release}"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        with self.assertRaisesRegex(
            ConfigError, "Invalid feed release format: abc"
        ):
            Config.load(path_mock)

        feed_release = "a.b"
        content = f"""[greenbone-feed-sync]
feed-release = "{feed_release}"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        with self.assertRaisesRegex(
            ConfigError, "Invalid feed release format: a.b"
        ):
            Config.load(path_mock)

    def test_invalid_toml(self):
        content = "This is not TOML"
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content
        path_mock.absolute.return_value = "/foo/bar.toml"

        with self.assertRaisesRegex(
            ConfigFileError,
            "Can't load config file. /foo/bar.toml is not a valid TOML file.",
        ):
            Config.load(path_mock)

    def test_load_ioerror(self):
        with self.assertRaisesRegex(
            ConfigFileError,
            r"Can't load config file .*foo\.toml\. Error was .*",
        ):
            Config.load(Path("foo.toml"))

    def test_getitem(self):
        content = """[greenbone-feed-sync]
no-wait = false
compression-level = 7
private-directory = "private"
"""
        path_mock = MagicMock(spec=Path)
        path_mock.read_text.return_value = content

        values = Config.load(path_mock)

        self.assertFalse(values["no-wait"])
        self.assertEqual(values["compression-level"], 7)
        self.assertEqual(values["private-directory"], Path("private"))

    def test_setitem(self):
        config = Config()

        config["no-wait"] = True
        config["compression-level"] = 7
        config["private-directory"] = "private"

        self.assertTrue(config["no-wait"])
        self.assertEqual(config["compression-level"], 7)
        self.assertEqual(config["private-directory"], "private")


class EnterpriseSettingsTestCase(unittest.TestCase):
    def test_from_key(self):
        content = """a_user@some.feed.server:/feed/
Lorem ipsum dolor sit amet,
consetetur sadipscing elitr,
sed diam nonumy eirmod tempor
"""
        with temp_file(content=content, name="enterprise.key") as f:
            settings = EnterpriseSettings.from_key(f)

            self.assertEqual(settings.key, f)
            self.assertEqual(settings.host, "some.feed.server")
            self.assertEqual(settings.user, "a_user")

    def test_feed_url(self):
        key = Path("/tmp/some.key")
        settings = EnterpriseSettings("foo", "some.server", key)

        self.assertEqual(
            settings.feed_url(), "ssh://foo@some.server/enterprise"
        )



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import asyncio
import subprocess
import sys
from dataclasses import dataclass
from typing import Iterable, NoReturn

from rich.console import Console

from greenbone.feed.sync.config import DEFAULT_VERBOSITY
from greenbone.feed.sync.errors import GreenboneFeedSyncError, RsyncError
from greenbone.feed.sync.helper import (
    Spinner,
    change_user_and_group,
    flock_wait,
    is_root,
)
from greenbone.feed.sync.parser import CliParser
from greenbone.feed.sync.rsync import Rsync

__all__ = ("main",)


@dataclass
class Sync:
    """
    Class to store sync information
    """

    name: str
    types: Iterable[str]
    url: str
    destination: str


@dataclass
class SyncList:
    """
    Class to store a list of sync information
    """

    lock_file: str
    syncs: Iterable[Sync]


def filter_syncs(lock_file: str, feed_type: str, *syncs: Sync) -> SyncList:
    """
    Create a list of syncs which only match to the feed type
    """
    return SyncList(
        lock_file=lock_file,
        syncs=[sync for sync in syncs if feed_type in sync.types],
    )


def do_selftest() -> None:
    """
    Check for rsync command.
    """
    try:
        subprocess.run(
            ["rsync", "--help"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
    except (PermissionError, FileNotFoundError, subprocess.CalledProcessError):
        raise GreenboneFeedSyncError(
            "The rsync binary could not be found."
        ) from None


async def feed_sync(console: Console, error_console: Console) -> int:
    """
    Sync the feeds
    """
    parser = CliParser()
    args = parser.parse_arguments()

    do_selftest()

    if args.selftest:
        return 0

    if args.quiet:
        verbose = 0
    else:
        verbose = DEFAULT_VERBOSITY if args.verbose is None else args.verbose

    if is_root():
        if verbose >= 1:
            console.print(
                f"Running as root. Switching to user '{args.user}' and "
                f"group '{args.group}'."
            )
            change_user_and_group(args.user, args.group)

    rsync = Rsync(
        private_subdir=args.private_directory,
        verbose=verbose >= 3,
        compression_level=args.compression_level,
        ssh_key=args.greenbone_enterprise_feed_key,
    )

    openvas_syncs = filter_syncs(
        args.openvas_lock_file,
        args.type,
        Sync(
            name="Notus files",
            types=("notus", "nvt", "all"),
            url=args.notus_url,
            destination=args.notus_destination,
        ),
        Sync(
            name="NASL files",
            types=("nasl", "nvt", "all"),
            url=args.nasl_url,
            destination=args.nasl_destination,
        ),
    )
    gvmd_syncs = filter_syncs(
        args.gvmd_lock_file,
        args.type,
        Sync(
            name="SCAP data",
            types=("scap", "all"),
            url=args.scap_data_url,
            destination=args.scap_data_destination,
        ),
        Sync(
            name="CERT-Bund data",
            types=("cert", "all"),
            url=args.cert_data_url,
            destination=args.cert_data_destination,
        ),
        Sync(
            name="gvmd data",
            types=("gvmd-data", "all"),
            url=args.gvmd_data_url,
            destination=args.gvmd_data_destination,
        ),
        Sync(
            name="report formats",
            types=("report-format"),
            url=args.report_formats_url,
            destination=args.report_formats_destination,
        ),
        Sync(
            name="scan configs",
            types=("scan-config"),
            url=args.scan_configs_url,
            destination=args.scan_configs_destination,
        ),
        Sync(
            name="port lists",
            types=("port-list"),
            url=args.port_lists_url,
            destination=args.port_lists_destination,
        ),
    )

    has_error = False
    wait_interval = None if args.no_wait else args.wait_interval

    for sync_list in (openvas_syncs, gvmd_syncs):
        if not sync_list.syncs:
            continue

        async with flock_wait(
            sync_list.lock_file,
            console=console if verbose else None,
            wait_interval=wait_interval,
        ):
            for sync in sync_list.syncs:
                try:
                    rsync_coro = rsync.sync(
                        url=sync.url, destination=sync.destination
                    )
                    if verbose >= 3:
                        console.print(
                            f"Downloading {sync.name} from {sync.url} to "
                            f"{sync.destination}"
                        )
                        await rsync_coro
                        # add newline after rsync
                        console.print()
                    elif verbose >= 1:
                        with Spinner(
                            console,
                            f"Downloading {sync.name} from {sync.url} to "
                            f"{sync.destination}",
                        ):
                            await rsync_coro
                    else:
                        await rsync_coro
                except RsyncError as e:
                    has_error = True
                    error_console.print(e.stderr)
                    if args.fail_fast:
                        return 1

        if verbose >= 2:
            # add newline for grouping lock
            console.print()

    return 1 if has_error else 0


def main() -> NoReturn:
    """
    Main CLI function
    """
    console = Console()
    error_console = Console(stderr=True)

    try:
        sys.exit(asyncio.run(feed_sync(console, error_console)))
    except GreenboneFeedSyncError as e:
        error_console.print(f"[red]❌[/red]Error: {e}")
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Optional,
    Protocol,
    TypeVar,
    Union,
)
from urllib.parse import urlsplit

from greenbone.feed.sync.errors import ConfigError, ConfigFileError
from greenbone.feed.sync.helper import DEFAULT_FLOCK_WAIT_INTERVAL
from greenbone.feed.sync.rsync import (
    DEFAULT_RSYNC_COMPRESSION_LEVEL,
    DEFAULT_RSYNC_URL,
)

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore[no-redef]


def maybe_int(value: Optional[str]) -> Union[int, str, None]:
    """
    Convert string into int if possible
    """
    try:
        value = int(value)  # type: ignore[arg-type,assignment]
    except ValueError:
        pass

    return value


DEFAULT_FEED_RELEASE = "24.10"

DEFAULT_DESTINATION_PREFIX = "/var/lib/"

DEFAULT_NASL_PATH = "openvas/plugins"
DEFAULT_NOTUS_PATH = "notus"
DEFAULT_SCAP_DATA_PATH = "gvm/scap-data"
DEFAULT_CERT_DATA_PATH = "gvm/cert-data"

DEFAULT_GVMD_LOCK_FILE_PATH = "gvm/feed-update.lock"
DEFAULT_OPENVAS_LOCK_FILE_PATH = "openvas/feed-update.lock"

DEFAULT_CONFIG_FILE = "/etc/gvm/greenbone-feed-sync.toml"
DEFAULT_USER_CONFIG_FILE = "~/.config/greenbone-feed-sync.toml"

DEFAULT_ENTERPRISE_KEY_PATH = "/etc/gvm/greenbone-enterprise-feed-key"

DEFAULT_GROUP = "gvm"
DEFAULT_USER = "gvm"

DEFAULT_VERBOSITY = 2

T = TypeVar("T")
ValuesDict = dict[str, Any]
DefaultValueCallable = Callable[[ValuesDict], Any]
ValueTypeCallable = Callable[[Any], T]


def resolve_gvmd_data_destination(values: ValuesDict) -> str:
    path = "gvm/data-objects/gvmd"
    feed_release: str = values.get("feed-release")  # type: ignore[assignment]
    try:
        str_major, str_minor = feed_release.split(".")[:2]
        major, minor = int(str_major), int(str_minor)
    except ValueError as e:
        raise ConfigError(f"Invalid feed release format: {feed_release}") from e

    return (
        f"{values['destination-prefix']}/{path}"
        if major >= 24 and minor >= 10
        else f"{values['destination-prefix']}/{path}/{feed_release}"
    )


@dataclass
class Setting(Generic[T]):
    config_key: str
    environment_key: str
    default_value: Union[str, int, bool, None]
    value_type: ValueTypeCallable[T]

    def resolve(self, values: ValuesDict) -> Optional[T]:
        value: Any
        if self.environment_key in os.environ:
            value = os.environ.get(self.environment_key)
        elif self.config_key in values:
            value = values.get(self.config_key)
        else:
            value = self.default_value

        return None if value is None else self.value_type(value)


@dataclass
class DependentSetting(Generic[T]):
    config_key: str
    environment_key: str
    default_value: DefaultValueCallable
    value_type: ValueTypeCallable[T]

    def resolve(self, values: ValuesDict) -> Optional[T]:
        if self.environment_key in os.environ:
            value = os.environ.get(self.environment_key)
        elif self.config_key in values:
            value = values.get(self.config_key)
        else:
            value = self.default_value(values)

        return None if value is None else self.value_type(value)


@dataclass
class EnterpriseSettings:
    user: Optional[str]
    host: Optional[str]
    key: Path

    @classmethod
    def from_key(cls, enterprise_key: Path) -> "EnterpriseSettings":
        with enterprise_key.open("r", encoding="utf8", errors="ignore") as f:
            line = f.readline()

        url = urlsplit(line)
        if not url.scheme:
            # ensure that url gets splitted correctly if line doesn't contain
            # an url scheme (which is the default)
            url = urlsplit(f"//{line}")

        return EnterpriseSettings(url.username, url.hostname, enterprise_key)

    def feed_url(self) -> str:
        return f"ssh://{self.user}@{self.host}/enterprise"


_SETTINGS = (
    Setting(
        "destination-prefix",
        "GREENBONE_FEED_SYNC_DESTINATION_PREFIX",
        DEFAULT_DESTINATION_PREFIX,
        Path,
    ),
    Setting("feed-url", "GREENBONE_FEED_SYNC_URL", DEFAULT_RSYNC_URL, str),
    Setting(
        "wait-interval",
        "GREENBONE_FEED_SYNC_LOCK_WAIT_INTERVAL",
        DEFAULT_FLOCK_WAIT_INTERVAL,
        int,
    ),
    Setting("no-wait", "GREENBONE_FEED_SYNC_NO_WAIT", False, bool),
    Setting(
        "compression-level",
        "GREENBONE_FEED_SYNC_COMPRESSION_LEVEL",
        DEFAULT_RSYNC_COMPRESSION_LEVEL,
        int,
    ),
    Setting(
        "private-directory", "GREENBONE_FEED_SYNC_PRIVATE_DIRECTORY", None, Path
    ),
    Setting("verbose", "GREENBONE_FEED_SYNC_VERBOSE", None, int),
    Setting("fail-fast", "GREENBONE_FEED_SYNC_FAIL_FAST", False, bool),
    Setting("rsync-timeout", "GREENBONE_FEED_SYNC_RSYNC_TIMEOUT", None, int),
    Setting("group", "GREENBONE_FEED_SYNC_GROUP", DEFAULT_GROUP, maybe_int),
    Setting("user", "GREENBONE_FEED_SYNC_USER", DEFAULT_USER, maybe_int),
    Setting(
        "greenbone-enterprise-feed-key",
        "GREENBONE_FEED_SYNC_ENTERPRISE_FEED_KEY",
        DEFAULT_ENTERPRISE_KEY_PATH,
        Path,
    ),
    Setting(
        "feed-release",
        "GREENBONE_FEED_SYNC_FEED_RELEASE",
        DEFAULT_FEED_RELEASE,
        str,
    ),
)


_DEPENDENT_SETTINGS = (
    DependentSetting(
        "gvmd-data-destination",
        "GREENBONE_FEED_SYNC_GVMD_DATA_DESTINATION",
        resolve_gvmd_data_destination,
        Path,
    ),
    DependentSetting(
        "gvmd-data-url",
        "GREENBONE_FEED_SYNC_GVMD_DATA_URL",
        lambda values: f"{values['feed-url']}/data-feed/{values['feed-release']}/",
        str,
    ),
    DependentSetting(
        "notus-destination",
        "GREENBONE_FEED_SYNC_NOTUS_DESTINATION",
        lambda values: f"{values['destination-prefix']}/{DEFAULT_NOTUS_PATH}",
        Path,
    ),
    DependentSetting(
        "notus-url",
        "GREENBONE_FEED_SYNC_NOTUS_URL",
        lambda values: f"{values['feed-url']}/vulnerability-feed/{values['feed-release']}/vt-data/notus/",
        str,
    ),
    DependentSetting(
        "nasl-destination",
        "GREENBONE_FEED_SYNC_NASL_DESTINATION",
        lambda values: f"{values['destination-prefix']}/{DEFAULT_NASL_PATH}",
        Path,
    ),
    DependentSetting(
        "nasl-url",
        "GREENBONE_FEED_SYNC_NASL_URL",
        lambda values: f"{values['feed-url']}/vulnerability-feed/{values['feed-release']}/vt-data/nasl/",
        str,
    ),
    DependentSetting(
        "scap-data-destination",
        "GREENBONE_FEED_SYNC_SCAP_DATA_DESTINATION",
        lambda values: f"{values['destination-prefix']}/{DEFAULT_SCAP_DATA_PATH}",  # noqa: E501
        Path,
    ),
    DependentSetting(
        "scap-data-url",
        "GREENBONE_FEED_SYNC_SCAP_DATA_URL",
        lambda values: f"{values['feed-url']}/vulnerability-feed/{values['feed-release']}/scap-data/",
        str,
    ),
    DependentSetting(
        "cert-data-destination",
        "GREENBONE_FEED_SYNC_CERT_DATA_DESTINATION",
        lambda values: f"{values['destination-prefix']}/{DEFAULT_CERT_DATA_PATH}",  # noqa: E501
        Path,
    ),
    DependentSetting(
        "cert-data-url",
        "GREENBONE_FEED_SYNC_CERT_DATA_URL",
        lambda values: f"{values['feed-url']}/vulnerability-feed/{values['feed-release']}/cert-data/",
        str,
    ),
    DependentSetting(
        "report-formats-destination",
        "GREENBONE_FEED_SYNC_REPORT_FORMATS_DESTINATION",
        lambda values: f"{values['gvmd-data-destination']}/report-formats",
        Path,
    ),
    DependentSetting(
        "report-formats-url",
        "GREENBONE_FEED_SYNC_REPORT_FORMATS_URL",
        lambda values: f"{values['feed-url']}/data-feed/{values['feed-release']}/report-formats/",  # noqa: E501
        str,
    ),
    DependentSetting(
        "scan-configs-destination",
        "GREENBONE_FEED_SYNC_SCAN_CONFIGS_DESTINATION",
        lambda values: f"{values['gvmd-data-destination']}/scan-configs",
        Path,
    ),
    DependentSetting(
        "scan-configs-url",
        "GREENBONE_FEED_SYNC_SCAN_CONFIGS_URL",
        lambda values: f"{values['feed-url']}/data-feed/{values['feed-release']}/scan-configs/",  # noqa: E501
        str,
    ),
    DependentSetting(
        "port-lists-destination",
        "GREENBONE_FEED_SYNC_PORT_LISTS_DESTINATION",
        lambda values: f"{values['gvmd-data-destination']}/port-lists",
        Path,
    ),
    DependentSetting(
        "port-lists-url",
        "GREENBONE_FEED_SYNC_PORT_LISTS_URL",
        lambda values: f"{values['feed-url']}/data-feed/{values['feed-release']}/port-lists/",  # noqa: E501
        str,
    ),
    DependentSetting(
        "gvmd-lock-file",
        "GREENBONE_FEED_SYNC_GVMD_LOCK_FILE",
        lambda values: f"{values['destination-prefix']}/{DEFAULT_GVMD_LOCK_FILE_PATH}",  # noqa: E501
        Path,
    ),
    DependentSetting(
        "openvas-lock-file",
        "GREENBONE_FEED_SYNC_OPENVAS_LOCK_FILE",
        lambda values: f"{values['destination-prefix']}/{DEFAULT_OPENVAS_LOCK_FILE_PATH}",  # noqa: E501
        Path,
    ),
)


class ConfigDict(Protocol):
    def items(self) -> Iterable[tuple[str, Any]]: ...

    def __getitem__(self, key: str) -> Any: ...


class Config:
    """
    A class to load configuration values from the environment, config file and
    defaults.
    """

    def __init__(self) -> None:
        self._config: ValuesDict = {}

    def load_from_config_file(self, config_file: Path) -> None:
        try:
            content = config_file.read_text(encoding="utf-8")
            config_data = tomllib.loads(content)
            self._config = config_data.get("greenbone-feed-sync", {})
        except IOError as e:
            raise ConfigFileError(
                f"Can't load config file {config_file.absolute()}. "
                f"Error was {e}."
            ) from e
        except tomllib.TOMLDecodeError as e:
            raise ConfigFileError(
                f"Can't load config file. {config_file.absolute()} is not "
                "a valid TOML file."
            ) from e

    def apply_settings(self) -> None:
        for setting in _SETTINGS:
            self._config[setting.config_key] = setting.resolve(self._config)

    def apply_dependent_settings(self) -> None:
        for setting in _DEPENDENT_SETTINGS:
            self._config[setting.config_key] = setting.resolve(self._config)

    @classmethod
    def load(cls, config_file: Optional[Path] = None) -> "Config":
        """
        Load config values from config_file and apply all settings
        """
        config = cls()

        if config_file:
            config.load_from_config_file(config_file)

        config.apply_settings()
        config.apply_dependent_settings()

        return config

    def items(self) -> Iterable[tuple[str, Any]]:
        return self._config.items()

    def __getitem__(self, key: str) -> Any:
        return self._config[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._config[key] = value

    def __len__(self) -> int:
        return len(self._config)



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import asyncio
import errno
import fcntl
import os
import shutil
from contextlib import asynccontextmanager
from pathlib import Path
from types import TracebackType
from typing import AsyncGenerator, Optional, Type, Union

from rich.console import Console
from rich.live import Live
from rich.spinner import Spinner as RichSpinner

from greenbone.feed.sync.errors import FileLockingError, GreenboneFeedSyncError

DEFAULT_FLOCK_WAIT_INTERVAL = 5  # in seconds


def is_root() -> bool:
    """
    Checks if the current user is root
    """
    return os.geteuid() == 0


@asynccontextmanager
async def flock_wait(
    path: Union[str, Path],
    *,
    console: Optional[Console] = None,
    wait_interval: Optional[Union[int, float]] = DEFAULT_FLOCK_WAIT_INTERVAL,
) -> AsyncGenerator[None, None]:
    """
    Try to lock a file and wait if it is already locked

    Arguments:
        path: File to lock
        console: A console to print messages to or None to keep the function
            quiet.
        wait_interval: Time to wait in seconds after failed lock attempt before
            re-trying to lock the file. Set to None to raise a FileLockingError
            instead of re-trying to acquire the lock. Default is 5 seconds.
    """
    # ensure path is a Path
    path = Path(path)

    # ensure parent directories exist
    try:
        path.parent.mkdir(parents=True, exist_ok=True, mode=0o770)
    except OSError as e:
        raise FileLockingError(
            f"Could not create parent directories for {path}"
        ) from e

    try:
        with path.open("w", encoding="utf8") as fd0:
            has_lock = False
            while not has_lock:
                try:
                    if console:
                        console.print(
                            f"Trying to acquire lock on {path.absolute()}"
                        )

                    fcntl.flock(fd0, fcntl.LOCK_EX | fcntl.LOCK_NB)

                    if console:
                        console.print(f"Acquired lock on {path.absolute()}")

                    has_lock = True
                    path.chmod(mode=0o660)
                except OSError as e:
                    if e.errno in (errno.EAGAIN, errno.EACCES):
                        if wait_interval is None:
                            raise FileLockingError(
                                f"{path.absolute()} is locked. Another process "
                                "related to the feed update may already running."
                            ) from None

                        if console:
                            console.print(
                                f"{path.absolute()} is locked by another process. "
                                f"Waiting {wait_interval} seconds before next try."
                            )
                        await asyncio.sleep(wait_interval)
                    else:
                        raise

            try:
                yield
            finally:
                try:
                    # free the lock
                    if console:
                        console.print(f"Releasing lock on {path.absolute()}")

                    fcntl.flock(fd0, fcntl.LOCK_UN)
                except OSError:
                    pass
    except PermissionError as e:
        raise FileLockingError(
            "Permission error while trying to open the lock file "
            f"{path.absolute()}"
        ) from e


class Spinner:
    def __init__(self, console: Console, status: str) -> None:
        self._spinner = RichSpinner(
            "dots", text=status, style="status.spinner", speed=1.0
        )
        self._live = Live(
            self._spinner,
            console=console,
            refresh_per_second=12.5,
            transient=False,
        )

    def __enter__(self) -> "Spinner":
        self._live.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._live.stop()


def change_user_and_group(
    user: Union[str, int], group: Union[str, int]
) -> None:
    """
    Change effective user or group of the current running process

    Args:
        user: User name or ID
        group: Group name or ID
    """
    if isinstance(user, str):
        user_id = shutil._get_uid(  # type: ignore[attr-defined] # pylint: disable=protected-access # noqa: E501
            user
        )
        if user_id is None:
            raise GreenboneFeedSyncError(
                f"Can't run as user '{user}'. User '{user}' is unknown."
            )
        user = user_id
    if isinstance(group, str):
        group_id = shutil._get_gid(group)  # type: ignore[attr-defined] # pylint: disable=protected-access # noqa: E501
        if group_id is None:
            raise GreenboneFeedSyncError(
                f"Can't run as group '{group}'. Group '{group}' is unknown."
            )
        group = group_id

    os.setegid(group)  # type: ignore[arg-type]
    os.seteuid(user)  # type: ignore[arg-type]



# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import sys
import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from pontos.testing import temp_directory

from greenbone.feed.sync.config import DEFAULT_FEED_RELEASE
from greenbone.feed.sync.errors import GreenboneFeedSyncError, RsyncError
from greenbone.feed.sync.main import (
    Sync,
    do_selftest,
    feed_sync,
    filter_syncs,
    main,
)


class FilterSyncsTestCase(unittest.TestCase):
    def test_filter_syncs(self):
        sync_a = Sync(name="a", types=["foo", "bar"], url="a", destination="a")
        sync_b = Sync(name="b", types=["foo", "baz"], url="b", destination="b")
        sync_c = Sync(name="c", types=["bar", "baz"], url="c", destination="c")

        sync_list = filter_syncs(
            "file.lock",
            "foo",
            sync_a,
            sync_b,
            sync_c,
        )

        self.assertEqual(len(sync_list.syncs), 2)
        self.assertEqual(sync_list.lock_file, "file.lock")

        self.assertEqual(sync_list.syncs[0], sync_a)
        self.assertEqual(sync_list.syncs[1], sync_b)


class DoSelftestTestCase(unittest.TestCase):
    @patch("greenbone.feed.sync.main.subprocess.run")
    def test_do_selftest_success(self, mock_subprocess_run: MagicMock):
        mock_subprocess_run.side_effect = [""]
        do_selftest()

    @patch("greenbone.feed.sync.main.subprocess.run")
    def test_do_selftest_rsync_fail(self, mock_subprocess_run: MagicMock):
        mock_subprocess_run.side_effect = [PermissionError]
        with self.assertRaisesRegex(
            GreenboneFeedSyncError, "The rsync binary could not be found."
        ):
            do_selftest()


class FeedSyncTestCase(unittest.IsolatedAsyncioTestCase):
    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    @patch("greenbone.feed.sync.main.change_user_and_group", autospec=True)
    @patch("greenbone.feed.sync.main.is_root", autospec=True)
    async def test_do_not_run_as_root(
        self,
        is_root_mock: MagicMock,
        change_user_mock: MagicMock,
        rsync_mock: MagicMock,
    ):
        is_root_mock.return_value = True
        console = MagicMock()
        rsync_mock_instance = rsync_mock.return_value

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                [
                    "greenbone-feed-sync",
                    "--type",
                    "nvt",
                ],
            ),
        ):
            ret = await feed_sync(console=console, error_console=console)
            self.assertEqual(ret, 0)

        change_user_mock.assert_called_once_with("gvm", "gvm")
        rsync_mock.assert_called_once_with(
            private_subdir=None,
            verbose=False,
            compression_level=9,
            ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
        )
        console.print.assert_has_calls(
            [
                call(
                    "Trying to acquire lock on "
                    f"{temp_dir}/openvas/feed-update.lock"
                ),
                call(f"Acquired lock on {temp_dir}/openvas/feed-update.lock"),
                call(f"Releasing lock on {temp_dir}/openvas/feed-update.lock"),
                call(),
            ]
        )

        rsync_mock_instance.sync.assert_has_awaits(
            [
                call(
                    url="rsync://feed.community.greenbone.net/community/"
                    f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                    destination=temp_dir / "notus",
                ),
                call(
                    url="rsync://feed.community.greenbone.net/community/"
                    f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
                    destination=temp_dir / "openvas/plugins",
                ),
            ]
        )

    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    async def test_sync_nvts(self, rsync_mock: MagicMock):
        console = MagicMock()
        rsync_mock_instance = rsync_mock.return_value

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                [
                    "greenbone-feed-sync",
                    "--type",
                    "nvt",
                ],
            ),
        ):
            ret = await feed_sync(console=console, error_console=console)
            self.assertEqual(ret, 0)

            rsync_mock.assert_called_once_with(
                private_subdir=None,
                verbose=False,
                compression_level=9,
                ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
            )
            console.print.assert_has_calls(
                [
                    call(
                        "Trying to acquire lock on "
                        f"{temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Acquired lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Releasing lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(),
                ]
            )

            rsync_mock_instance.sync.assert_has_awaits(
                [
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                        destination=temp_dir / "notus",
                    ),
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
                        destination=temp_dir / "openvas/plugins",
                    ),
                ]
            )

    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    async def test_sync_nvts_verbose(self, rsync_mock: MagicMock):
        console = MagicMock()
        rsync_mock_instance = rsync_mock.return_value

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                ["greenbone-feed-sync", "--type", "nvt", "-vvv"],
            ),
        ):
            ret = await feed_sync(console=console, error_console=console)
            self.assertEqual(ret, 0)

            rsync_mock.assert_called_once_with(
                private_subdir=None,
                verbose=True,
                compression_level=9,
                ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
            )
            console.print.assert_has_calls(
                [
                    call(
                        "Trying to acquire lock on "
                        f"{temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Acquired lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        "Downloading Notus files from "
                        "rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/ "
                        f"to {temp_dir}/notus"
                    ),
                    call(),
                    call(
                        "Downloading NASL files from "
                        "rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/ "
                        f"to {temp_dir}/openvas/plugins"
                    ),
                    call(),
                    call(
                        f"Releasing lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(),
                ]
            )

            rsync_mock_instance.sync.assert_has_awaits(
                [
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                        destination=temp_dir / "notus",
                    ),
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
                        destination=temp_dir / "openvas/plugins",
                    ),
                ]
            )

    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    async def test_sync_nvts_quiet(self, rsync_mock: MagicMock):
        console = MagicMock()
        rsync_mock_instance = rsync_mock.return_value

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                ["greenbone-feed-sync", "--type", "nvt", "--quiet"],
            ),
        ):
            ret = await feed_sync(console=console, error_console=console)
            self.assertEqual(ret, 0)

            rsync_mock.assert_called_once_with(
                private_subdir=None,
                verbose=False,
                compression_level=9,
                ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
            )
            console.print.assert_not_called()

            rsync_mock_instance.sync.assert_has_awaits(
                [
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                        destination=temp_dir / "notus",
                    ),
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
                        destination=temp_dir / "openvas/plugins",
                    ),
                ]
            )

    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    async def test_sync_nvts_rsync_error(self, rsync_mock: MagicMock):
        console = MagicMock()
        rsync_mock_instance = rsync_mock.return_value
        rsync_mock_instance.sync.side_effect = RsyncError(
            2, [], b"An rsync error"
        )

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                ["greenbone-feed-sync", "--type", "nvt", "--fail-fast"],
            ),
        ):
            ret = await feed_sync(console=console, error_console=console)
            self.assertEqual(ret, 1)

            rsync_mock.assert_called_once_with(
                private_subdir=None,
                verbose=False,
                compression_level=9,
                ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
            )
            console.print.assert_has_calls(
                [
                    call(
                        "Trying to acquire lock on "
                        f"{temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Acquired lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call("An rsync error"),
                    call(
                        f"Releasing lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                ]
            )

            rsync_mock_instance.sync.assert_has_awaits(
                [
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                        destination=temp_dir / "notus",
                    ),
                ]
            )


class MainFunctionTestCase(unittest.TestCase):
    @patch("greenbone.feed.sync.main.Console")
    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    def test_sync_nvts(self, rsync_mock: MagicMock, console_mock: MagicMock):
        rsync_mock_instance = rsync_mock.return_value
        console_mock_instance = console_mock.return_value

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                [
                    "greenbone-feed-sync",
                    "--type",
                    "nvt",
                ],
            ),
        ):
            with self.assertRaises(SystemExit) as cm:
                main()

            self.assertEqual(cm.exception.code, 0)

            rsync_mock.assert_called_once_with(
                private_subdir=None,
                verbose=False,
                compression_level=9,
                ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
            )
            console_mock_instance.print.assert_has_calls(
                [
                    call(
                        "Trying to acquire lock on "
                        f"{temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Acquired lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Releasing lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(),
                ]
            )

            rsync_mock_instance.sync.assert_has_awaits(
                [
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                        destination=temp_dir / "notus",
                    ),
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
                        destination=temp_dir / "openvas/plugins",
                    ),
                ]
            )

    @patch("greenbone.feed.sync.main.Console")
    @patch("greenbone.feed.sync.main.Rsync", autospec=True)
    def test_sync_nvts_error(
        self, rsync_mock: MagicMock, console_mock: MagicMock
    ):
        rsync_mock_instance = rsync_mock.return_value
        console_mock_instance = console_mock.return_value
        rsync_mock_instance.sync.side_effect = GreenboneFeedSyncError(
            "An error"
        )

        with (
            temp_directory() as temp_dir,
            patch.dict(
                "os.environ",
                {"GREENBONE_FEED_SYNC_DESTINATION_PREFIX": str(temp_dir)},
            ),
            patch.object(
                sys,
                "argv",
                ["greenbone-feed-sync", "--type", "nvt", "--fail-fast"],
            ),
        ):
            with self.assertRaises(SystemExit) as cm:
                main()

            self.assertEqual(cm.exception.code, 1)

            rsync_mock.assert_called_once_with(
                private_subdir=None,
                verbose=False,
                compression_level=9,
                ssh_key=Path("/etc/gvm/greenbone-enterprise-feed-key"),
            )
            console_mock_instance.print.assert_has_calls(
                [
                    call(
                        "Trying to acquire lock on "
                        f"{temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Acquired lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call(
                        f"Releasing lock on {temp_dir}/openvas/feed-update.lock"
                    ),
                    call("[red]❌[/red]Error: An error"),
                ]
            )

            rsync_mock_instance.sync.assert_has_awaits(
                [
                    call(
                        url="rsync://feed.community.greenbone.net/community/"
                        f"vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
                        destination=temp_dir / "notus",
                    ),
                ]
            )



# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import asyncio
import unittest
from asyncio.subprocess import Process
from pathlib import Path
from unittest.mock import AsyncMock, patch

from greenbone.feed.sync.errors import RsyncError
from greenbone.feed.sync.rsync import Rsync, exec_rsync


class RsyncTestCase(unittest.IsolatedAsyncioTestCase):
    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_defaults(self, exec_mock: AsyncMock):
        rsync = Rsync()
        await rsync.sync("rsync://foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "-q",
            "--compress-level=9",
            "--delete",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "rsync://foo.bar/baz",
            "/tmp/baz",
        )

    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_private_subdir(self, exec_mock: AsyncMock):
        rsync = Rsync(private_subdir="private")
        await rsync.sync("rsync://foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "-q",
            "--compress-level=9",
            "--delete",
            "--exclude",
            "private",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "rsync://foo.bar/baz",
            "/tmp/baz",
        )

    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_verbose(self, exec_mock: AsyncMock):
        rsync = Rsync(verbose=True)
        await rsync.sync("rsync://foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "-v",
            "--compress-level=9",
            "--delete",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "rsync://foo.bar/baz",
            "/tmp/baz",
        )

    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_compression_level(self, exec_mock: AsyncMock):
        rsync = Rsync(compression_level=1)
        await rsync.sync("rsync://foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "-q",
            "--compress-level=1",
            "--delete",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "rsync://foo.bar/baz",
            "/tmp/baz",
        )

    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_exclude(self, exec_mock: AsyncMock):
        rsync = Rsync(exclude=["foo", Path("exclude/this")])
        await rsync.sync("rsync://foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "-q",
            "--compress-level=9",
            "--delete",
            "--exclude",
            "foo",
            "--exclude",
            "exclude/this",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "rsync://foo.bar/baz",
            "/tmp/baz",
        )


class ExecRsyncTestCase(unittest.IsolatedAsyncioTestCase):
    @patch(
        "greenbone.feed.sync.rsync.asyncio.create_subprocess_exec",
        autospec=True,
    )
    async def test_failure(self, exec_mock: AsyncMock):
        process_mock = AsyncMock(spec=Process)
        process_mock.communicate.return_value = (None, b"An error occurred")
        process_mock.wait.return_value = 1
        exec_mock.return_value = process_mock

        with self.assertRaises(RsyncError) as cm:
            await exec_rsync("foo", "bar")

        exec_mock.assert_awaited_once_with(
            "rsync", "foo", "bar", stderr=asyncio.subprocess.PIPE
        )

        self.assertEqual(cm.exception.returncode, 1)
        self.assertEqual(cm.exception.stderr, "An error occurred")
        self.assertEqual(cm.exception.cmd, ["rsync", "foo", "bar"])
        self.assertIsNone(cm.exception.stout)
        self.assertEqual(
            str(cm.exception),
            "'rsync foo bar' returned non-zero exit status 1.",
        )

    @patch(
        "greenbone.feed.sync.rsync.asyncio.create_subprocess_exec",
        autospec=True,
    )
    async def test_success(self, exec_mock: AsyncMock):
        process_mock = AsyncMock(spec=Process)
        process_mock.communicate.return_value = (None, None)
        process_mock.wait.return_value = 0
        exec_mock.return_value = process_mock

        await exec_rsync("foo", "bar")

        exec_mock.assert_awaited_once_with(
            "rsync", "foo", "bar", stderr=asyncio.subprocess.PIPE
        )

    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_timeout(self, exec_mock: AsyncMock):
        rsync = Rsync(timeout=120)
        await rsync.sync("rsync://foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "--timeout=120",
            "-q",
            "--compress-level=9",
            "--delete",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "rsync://foo.bar/baz",
            "/tmp/baz",
        )

    @patch("greenbone.feed.sync.rsync.exec_rsync", autospec=True)
    async def test_rsync_with_ssh(self, exec_mock: AsyncMock):
        ssh_key = Path("/tmp/ssh.key")
        rsync = Rsync(ssh_key=ssh_key)
        await rsync.sync("ssh://user@foo.bar/baz", "/tmp/baz")

        exec_mock.assert_awaited_once_with(
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
            "-e",
            "ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -p 24 -i '/tmp/ssh.key'",  # pylint: disable=line-too-long # noqa: E501
            "-q",
            "--compress-level=9",
            "--delete",
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
            "--copy-unsafe-links",
            "--hard-links",
            "user@foo.bar:/baz",
            "/tmp/baz",
        )



# SPDX-FileCopyrightText: 2023-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=line-too-long, too-many-lines

import io
import sys
import unittest
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos.testing import temp_file

from greenbone.feed.sync.config import (
    DEFAULT_CONFIG_FILE,
    DEFAULT_DESTINATION_PREFIX,
    DEFAULT_ENTERPRISE_KEY_PATH,
    DEFAULT_FEED_RELEASE,
    DEFAULT_GVMD_LOCK_FILE_PATH,
    DEFAULT_OPENVAS_LOCK_FILE_PATH,
    DEFAULT_USER_CONFIG_FILE,
)
from greenbone.feed.sync.errors import ConfigFileError
from greenbone.feed.sync.helper import DEFAULT_FLOCK_WAIT_INTERVAL
from greenbone.feed.sync.parser import CliParser, feed_type
from greenbone.feed.sync.rsync import (
    DEFAULT_RSYNC_COMPRESSION_LEVEL,
    DEFAULT_RSYNC_URL,
)


class FeedTypeTestCase(unittest.TestCase):
    def test_nvt(self):
        self.assertEqual("nvt", feed_type("nvt"))
        self.assertEqual("nvt", feed_type("nvts"))
        self.assertEqual("nvt", feed_type("NVT"))
        self.assertEqual("nvt", feed_type("NVTS"))

    def test_report_format(self):
        self.assertEqual("report-format", feed_type("report-format"))
        self.assertEqual("report-format", feed_type("report-formats"))
        self.assertEqual("report-format", feed_type("report_format"))
        self.assertEqual("report-format", feed_type("report_formats"))
        self.assertEqual("report-format", feed_type("REPORT_FORMAT"))
        self.assertEqual("report-format", feed_type("REPORT_FORMATS"))
        self.assertEqual("report-format", feed_type("REPORT-FORMAT"))
        self.assertEqual("report-format", feed_type("REPORT-FORMATS"))

    def test_port_list(self):
        self.assertEqual("port-list", feed_type("port-list"))
        self.assertEqual("port-list", feed_type("port-lists"))
        self.assertEqual("port-list", feed_type("port_list"))
        self.assertEqual("port-list", feed_type("port_lists"))
        self.assertEqual("port-list", feed_type("PORT_LIST"))
        self.assertEqual("port-list", feed_type("PORT_LISTS"))
        self.assertEqual("port-list", feed_type("PORT-LIST"))
        self.assertEqual("port-list", feed_type("PORT-LISTS"))

    def test_scan_config(self):
        self.assertEqual("scan-config", feed_type("scan-config"))
        self.assertEqual("scan-config", feed_type("scan-configs"))
        self.assertEqual("scan-config", feed_type("scan_config"))
        self.assertEqual("scan-config", feed_type("scan_configs"))
        self.assertEqual("scan-config", feed_type("SCAN_CONFIG"))
        self.assertEqual("scan-config", feed_type("SCAN_CONFIGS"))
        self.assertEqual("scan-config", feed_type("SCAN-CONFIG"))
        self.assertEqual("scan-config", feed_type("SCAN-CONFIGS"))

    def test_gvmd_data(self):
        self.assertEqual("gvmd-data", feed_type("gvmd-data"))
        self.assertEqual("gvmd-data", feed_type("gvmd_data"))
        self.assertEqual("gvmd-data", feed_type("GVMD-DATA"))
        self.assertEqual("gvmd-data", feed_type("GVMD_DATA"))


class CliParserTestCase(unittest.TestCase):
    def test_output_group(self):
        parser = CliParser()

        with redirect_stderr(io.StringIO()) as f, self.assertRaises(SystemExit):
            parser.parse_arguments(["-vvv", "--quiet"])

        self.assertIn(
            "argument --quiet: not allowed with argument --verbose/-v",
            f.getvalue(),
        )

    def test_wait_group(self):
        parser = CliParser()

        with redirect_stderr(io.StringIO()) as f, self.assertRaises(SystemExit):
            parser.parse_arguments(["--no-wait", "--wait-interval", "20"])

        self.assertIn(
            "argument --wait-interval: not allowed with argument --no-wait",
            f.getvalue(),
        )

    def test_defaults(self):
        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.type, "all")
        self.assertEqual(
            args.destination_prefix, Path(DEFAULT_DESTINATION_PREFIX)
        )
        self.assertEqual(args.feed_url, DEFAULT_RSYNC_URL)
        self.assertEqual(
            args.gvmd_data_destination,
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "data-objects" / "gvmd",
        )
        self.assertEqual(
            args.gvmd_data_url,
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/",
        )
        self.assertEqual(
            args.notus_destination,
            Path(DEFAULT_DESTINATION_PREFIX) / "notus",
        )
        self.assertEqual(
            args.notus_url,
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
        )
        self.assertEqual(
            args.nasl_destination,
            Path(DEFAULT_DESTINATION_PREFIX) / "openvas" / "plugins",
        )
        self.assertEqual(
            args.nasl_url,
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
        )
        self.assertEqual(
            args.scap_data_destination,
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "scap-data",
        )
        self.assertEqual(
            args.scap_data_url,
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/scap-data/",
        )
        self.assertEqual(
            args.cert_data_destination,
            Path(DEFAULT_DESTINATION_PREFIX) / "gvm" / "cert-data",
        )
        self.assertEqual(
            args.cert_data_url,
            f"{DEFAULT_RSYNC_URL}/vulnerability-feed/{DEFAULT_FEED_RELEASE}/cert-data/",
        )
        self.assertEqual(
            args.report_formats_destination,
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "report-formats",
        )
        self.assertEqual(
            args.report_formats_url,
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/report-formats/",
        )
        self.assertEqual(
            args.scan_configs_destination,
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "scan-configs",
        )
        self.assertEqual(
            args.scan_configs_url,
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/scan-configs/",
        )
        self.assertEqual(
            args.port_lists_destination,
            Path(DEFAULT_DESTINATION_PREFIX)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "port-lists",
        )
        self.assertEqual(
            args.port_lists_url,
            f"{DEFAULT_RSYNC_URL}/data-feed/{DEFAULT_FEED_RELEASE}/port-lists/",
        )
        self.assertEqual(
            args.gvmd_lock_file,
            Path(DEFAULT_DESTINATION_PREFIX) / DEFAULT_GVMD_LOCK_FILE_PATH,
        )
        self.assertEqual(
            args.openvas_lock_file,
            Path(DEFAULT_DESTINATION_PREFIX) / DEFAULT_OPENVAS_LOCK_FILE_PATH,
        )
        self.assertEqual(args.wait_interval, DEFAULT_FLOCK_WAIT_INTERVAL)
        self.assertFalse(args.no_wait)
        self.assertEqual(
            args.compression_level, DEFAULT_RSYNC_COMPRESSION_LEVEL
        )
        self.assertIsNone(args.private_directory)
        self.assertIsNone(args.verbose)
        self.assertFalse(args.fail_fast)
        self.assertIsNone(args.rsync_timeout)
        self.assertEqual(
            args.greenbone_enterprise_feed_key,
            Path(DEFAULT_ENTERPRISE_KEY_PATH),
        )
        self.assertEqual(args.feed_release, DEFAULT_FEED_RELEASE)

    def test_help(self):
        parser = CliParser()

        with (
            redirect_stdout(io.StringIO()) as f,
            self.assertRaises(SystemExit) as cm,
        ):
            parser.parse_arguments(["--help"])

        self.assertEqual(cm.exception.code, 0)
        self.assertTrue(f.getvalue().startswith("usage: "))

    def test_verbose(self):
        parser = CliParser()

        args = parser.parse_arguments(["-v"])
        self.assertEqual(args.verbose, 1)

        args = parser.parse_arguments(["--verbose"])
        self.assertEqual(args.verbose, 1)

        args = parser.parse_arguments(["-vv"])
        self.assertEqual(args.verbose, 2)

        args = parser.parse_arguments(["-vvv"])
        self.assertEqual(args.verbose, 3)

        args = parser.parse_arguments(["-vvvv"])
        self.assertEqual(args.verbose, 4)

    def test_quiet(self):
        parser = CliParser()
        args = parser.parse_arguments(["--quiet"])
        self.assertTrue(args.quiet)

    def test_selftest(self):
        parser = CliParser()
        args = parser.parse_arguments(["--selftest"])
        self.assertTrue(args.selftest)

    @patch("greenbone.feed.sync.parser.Path")
    def test_use_default_config_files(self, path_mock):
        path_mock_instance = path_mock.return_value
        path_mock_instance.expanduser.return_value = path_mock_instance
        path_mock_instance.resolve.return_value = path_mock_instance
        path_mock_instance.exists.return_value = False

        parser = CliParser()
        parser.parse_arguments([])

        path_mock.assert_any_call(DEFAULT_USER_CONFIG_FILE)
        path_mock.assert_any_call(DEFAULT_CONFIG_FILE)

    def test_private_directory(self):
        parser = CliParser()
        args = parser.parse_arguments(["--private-directory", "foobar"])
        self.assertEqual(args.private_directory, Path("foobar"))

    def test_compression_level(self):
        parser = CliParser()
        args = parser.parse_arguments(["--compression-level", "0"])
        self.assertEqual(args.compression_level, 0)

        args = parser.parse_arguments(["--compression-level", "1"])
        self.assertEqual(args.compression_level, 1)

        args = parser.parse_arguments(["--compression-level", "2"])
        self.assertEqual(args.compression_level, 2)

        args = parser.parse_arguments(["--compression-level", "3"])
        self.assertEqual(args.compression_level, 3)

        args = parser.parse_arguments(["--compression-level", "4"])
        self.assertEqual(args.compression_level, 4)

        args = parser.parse_arguments(["--compression-level", "5"])
        self.assertEqual(args.compression_level, 5)

        args = parser.parse_arguments(["--compression-level", "6"])
        self.assertEqual(args.compression_level, 6)

        args = parser.parse_arguments(["--compression-level", "7"])
        self.assertEqual(args.compression_level, 7)

        args = parser.parse_arguments(["--compression-level", "8"])
        self.assertEqual(args.compression_level, 8)

        args = parser.parse_arguments(["--compression-level", "9"])
        self.assertEqual(args.compression_level, 9)

        with redirect_stderr(io.StringIO()) as f, self.assertRaises(SystemExit):
            parser.parse_arguments(["--compression-level", "10"])

        self.assertIn(
            "error: argument --compression-level: invalid choice:",
            f.getvalue(),
        )

    def test_gvmd_data_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--gvmd-data-destination", "foo/bar"])
        self.assertEqual(args.gvmd_data_destination, Path("foo/bar"))

    def test_gvmd_data_url(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--gvmd-data-url", "rsync://foo.bar/gvmd-data"]
        )
        self.assertEqual(args.gvmd_data_url, "rsync://foo.bar/gvmd-data")

    def test_notus_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--notus-destination", "foo/bar"])
        self.assertEqual(args.notus_destination, Path("foo/bar"))

    def test_notus_url(self):
        parser = CliParser()
        args = parser.parse_arguments(["--notus-url", "rsync://foo.bar/notus"])
        self.assertEqual(args.notus_url, "rsync://foo.bar/notus")

    def test_nasl_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--nasl-destination", "foo/bar"])
        self.assertEqual(args.nasl_destination, Path("foo/bar"))

    def test_nasl_url(self):
        parser = CliParser()
        args = parser.parse_arguments(["--nasl-url", "rsync://foo.bar/nasl"])
        self.assertEqual(args.nasl_url, "rsync://foo.bar/nasl")

    def test_scap_data_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--scap-data-destination", "foo/bar"])
        self.assertEqual(args.scap_data_destination, Path("foo/bar"))

    def test_scap_data_url(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--scap-data-url", "rsync://foo.bar/scap-data"]
        )
        self.assertEqual(args.scap_data_url, "rsync://foo.bar/scap-data")

    def test_cert_data_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--cert-data-destination", "foo/bar"])
        self.assertEqual(args.cert_data_destination, Path("foo/bar"))

    def test_cert_data_url(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--cert-data-url", "rsync://foo.bar/cert-data"]
        )
        self.assertEqual(args.cert_data_url, "rsync://foo.bar/cert-data")

    def test_report_formats_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--report-formats-destination", "foo/bar"]
        )
        self.assertEqual(args.report_formats_destination, Path("foo/bar"))

    def test_report_formats_url(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--report-formats-url", "rsync://foo.bar/report-formats"]
        )
        self.assertEqual(
            args.report_formats_url, "rsync://foo.bar/report-formats"
        )

    def test_scan_configs_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--scan-configs-destination", "foo/bar"])
        self.assertEqual(args.scan_configs_destination, Path("foo/bar"))

    def test_scan_configs_url(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--scan-configs-url", "rsync://foo.bar/scan-configs"]
        )
        self.assertEqual(args.scan_configs_url, "rsync://foo.bar/scan-configs")

    def test_port_lists_destination(self):
        parser = CliParser()
        args = parser.parse_arguments(["--port-lists-destination", "foo/bar"])
        self.assertEqual(args.port_lists_destination, Path("foo/bar"))

    def test_port_lists_url(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--port-lists-url", "rsync://foo.bar/port-lists"]
        )
        self.assertEqual(args.port_lists_url, "rsync://foo.bar/port-lists")

    def test_gvmd_lock_file(self):
        parser = CliParser()
        args = parser.parse_arguments(["--gvmd-lock-file", "/run/gvmd.lock"])
        self.assertEqual(args.gvmd_lock_file, Path("/run/gvmd.lock"))

    def test_openvas_lock_file(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--openvas-lock-file", "/run/openvas.lock"]
        )
        self.assertEqual(args.openvas_lock_file, Path("/run/openvas.lock"))

    def test_fail_fast(self):
        parser = CliParser()
        args = parser.parse_arguments(["--fail-fast"])
        self.assertTrue(args.fail_fast)

        args = parser.parse_arguments(["--failfast"])
        self.assertTrue(args.fail_fast)

    def test_no_wait(self):
        parser = CliParser()
        args = parser.parse_arguments(["--no-wait"])
        self.assertTrue(args.no_wait)

    def test_wait_interval(self):
        parser = CliParser()
        args = parser.parse_arguments(["--wait-interval", "100"])
        self.assertEqual(args.wait_interval, 100)

    def test_rsync_timeout(self):
        parser = CliParser()
        args = parser.parse_arguments(["--rsync-timeout", "120"])
        self.assertEqual(args.rsync_timeout, 120)

    def test_greenbone_enterprise_feed_key(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--greenbone-enterprise-feed-key", "/tmp/some.key"]
        )
        self.assertEqual(
            args.greenbone_enterprise_feed_key, Path("/tmp/some.key")
        )

    def test_other(self):
        parser = CliParser()

        with (
            redirect_stderr(io.StringIO()) as f,
            self.assertRaises(SystemExit) as cm,
        ):
            parser.parse_arguments(["--foo-bar", "10"])

        self.assertIn("error: unrecognized arguments: --foo-bar", f.getvalue())
        self.assertEqual(cm.exception.code, 2)

    @patch("greenbone.feed.sync.parser.Path")
    def test_config(self, path_mock: MagicMock):
        content = """[greenbone-feed-sync]
verbose = 3
feed-url = "rsync://foo.bar"
destination-prefix = "/usr/lib"
"""
        path_mock_instance = path_mock.return_value
        path_mock_instance.absolute.return_value = "/foo/bar/foo.toml"
        path_mock_instance.expanduser.return_value = path_mock_instance
        path_mock_instance.resolve.return_value = path_mock_instance
        path_mock_instance.exists.return_value = True
        path_mock_instance.read_text.return_value = content

        parser = CliParser()
        args = parser.parse_arguments(["--config", "foo.toml"])

        self.assertEqual(args.verbose, 3)
        self.assertEqual(args.feed_url, "rsync://foo.bar")
        self.assertEqual(args.destination_prefix, Path("/usr/lib"))

    @patch("greenbone.feed.sync.parser.Path")
    def test_load_from_default_config(self, path_mock: MagicMock):
        content = """[greenbone-feed-sync]
verbose = 3
feed-url = "rsync://foo.bar"
destination-prefix = "/usr/lib"
"""
        path_mock_instance = path_mock.return_value
        path_mock_instance.absolute.return_value = (
            "/foo/bar/.config/greenbone-feed-sync.toml"
        )
        path_mock_instance.expanduser.return_value = path_mock_instance
        path_mock_instance.resolve.return_value = path_mock_instance
        path_mock_instance.exists.return_value = True
        path_mock_instance.read_text.return_value = content

        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.verbose, 3)
        self.assertEqual(args.feed_url, "rsync://foo.bar")
        self.assertEqual(args.destination_prefix, Path("/usr/lib"))

    @patch.dict(
        "os.environ",
        {
            "GREENBONE_FEED_SYNC_DESTINATION_PREFIX": "/usr/lib",
            "GREENBONE_FEED_SYNC_URL": "rsync://foo.bar",
            "GREENBONE_FEED_SYNC_VERBOSE": "3",
        },
    )
    def test_from_environment(self):
        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.verbose, 3)
        self.assertEqual(args.feed_url, "rsync://foo.bar")
        self.assertEqual(args.destination_prefix, Path("/usr/lib"))

    @patch.dict(
        "os.environ",
        {
            "GREENBONE_FEED_SYNC_DESTINATION_PREFIX": "/usr/lib",
            "GREENBONE_FEED_SYNC_URL": "rsync://foo.bar",
            "GREENBONE_FEED_SYNC_VERBOSE": "3",
        },
    )
    @patch("greenbone.feed.sync.parser.Path")
    def test_environment_takes_precedence(self, path_mock):
        content = """[greenbone-feed-sync]
verbose = 9
feed-url = "rsync://lorem.ipsum"
destination-prefix = "/opt/lib"
wait-interval = 100
"""
        path_mock_instance = path_mock.return_value
        path_mock_instance.absolute.return_value = (
            "/foo/bar/.config/greenbone-feed-sync.toml"
        )
        path_mock_instance.expanduser.return_value = path_mock_instance
        path_mock_instance.resolve.return_value = path_mock_instance
        path_mock_instance.exists.return_value = True
        path_mock_instance.read_text.return_value = content

        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.verbose, 3)
        self.assertEqual(args.feed_url, "rsync://foo.bar")
        self.assertEqual(args.destination_prefix, Path("/usr/lib"))
        self.assertEqual(args.wait_interval, 100)

    @patch.dict(
        "os.environ",
        {
            "GREENBONE_FEED_SYNC_DESTINATION_PREFIX": "/usr/lib",
            "GREENBONE_FEED_SYNC_URL": "rsync://foo.bar",
            "GREENBONE_FEED_SYNC_VERBOSE": "3",
        },
    )
    @patch("greenbone.feed.sync.parser.Path")
    def test_argument_takes_precedence(self, path_mock):
        content = """[greenbone-feed-sync]
verbose = 9
feed-url = "rsync://lorem.ipsum"
destination-prefix = "/opt/lib"
wait-interval = 100
"""
        path_mock_instance = path_mock.return_value
        path_mock_instance.absolute.return_value = (
            "/foo/bar/.config/greenbone-feed-sync.toml"
        )
        path_mock_instance.expanduser.return_value = path_mock_instance
        path_mock_instance.resolve.return_value = path_mock_instance
        path_mock_instance.exists.return_value = True
        path_mock_instance.read_text.return_value = content

        parser = CliParser()
        args = parser.parse_arguments(
            [
                "--wait-interval",
                "90",
            ]
        )

        self.assertEqual(args.verbose, 3)
        self.assertEqual(args.feed_url, "rsync://foo.bar")
        self.assertEqual(args.destination_prefix, Path("/usr/lib"))
        self.assertEqual(args.wait_interval, 90)

    def test_config_file_not_exists(self):
        parser = CliParser()

        with self.assertRaisesRegex(
            ConfigFileError, "Config file foo.bar does not exist."
        ):
            parser.parse_arguments(["--config", "foo.bar"])

    def test_type(self):
        parser = CliParser()
        args = parser.parse_arguments(["--type", "nvt"])
        self.assertEqual(args.type, "nvt")
        args = parser.parse_arguments(["--type", "nvts"])
        self.assertEqual(args.type, "nvt")
        args = parser.parse_arguments(["--type", "NVT"])
        self.assertEqual(args.type, "nvt")
        args = parser.parse_arguments(["--type", "NVTS"])
        self.assertEqual(args.type, "nvt")

        args = parser.parse_arguments(["--type", "notus"])
        self.assertEqual(args.type, "notus")
        args = parser.parse_arguments(["--type", "NOTUS"])
        self.assertEqual(args.type, "notus")
        args = parser.parse_arguments(["--type", "NoTuS"])
        self.assertEqual(args.type, "notus")

        args = parser.parse_arguments(["--type", "nasl"])
        self.assertEqual(args.type, "nasl")
        args = parser.parse_arguments(["--type", "NASL"])
        self.assertEqual(args.type, "nasl")
        args = parser.parse_arguments(["--type", "NaSl"])
        self.assertEqual(args.type, "nasl")

        args = parser.parse_arguments(["--type", "scap"])
        self.assertEqual(args.type, "scap")
        args = parser.parse_arguments(["--type", "SCAP"])
        self.assertEqual(args.type, "scap")
        args = parser.parse_arguments(["--type", "ScAp"])
        self.assertEqual(args.type, "scap")

        args = parser.parse_arguments(["--type", "cert"])
        self.assertEqual(args.type, "cert")
        args = parser.parse_arguments(["--type", "CERT"])
        self.assertEqual(args.type, "cert")
        args = parser.parse_arguments(["--type", "CeRt"])
        self.assertEqual(args.type, "cert")

        args = parser.parse_arguments(["--type", "all"])
        self.assertEqual(args.type, "all")
        args = parser.parse_arguments(["--type", "ALL"])
        self.assertEqual(args.type, "all")
        args = parser.parse_arguments(["--type", "AlL"])
        self.assertEqual(args.type, "all")

        args = parser.parse_arguments(["--type", "report-format"])
        self.assertEqual(args.type, "report-format")
        args = parser.parse_arguments(["--type", "REPORT-FORMAT"])
        self.assertEqual(args.type, "report-format")
        args = parser.parse_arguments(["--type", "report_format"])
        self.assertEqual(args.type, "report-format")
        args = parser.parse_arguments(["--type", "REPORT_FORMAT"])
        self.assertEqual(args.type, "report-format")
        args = parser.parse_arguments(["--type", "Report-Format"])
        self.assertEqual(args.type, "report-format")
        args = parser.parse_arguments(["--type", "report-formats"])
        self.assertEqual(args.type, "report-format")
        args = parser.parse_arguments(["--type", "REPORT_FORMATS"])
        self.assertEqual(args.type, "report-format")

        args = parser.parse_arguments(["--type", "scan-config"])
        self.assertEqual(args.type, "scan-config")
        args = parser.parse_arguments(["--type", "SCAN-CONFIG"])
        self.assertEqual(args.type, "scan-config")
        args = parser.parse_arguments(["--type", "scan_config"])
        self.assertEqual(args.type, "scan-config")
        args = parser.parse_arguments(["--type", "SCAN_config"])
        self.assertEqual(args.type, "scan-config")
        args = parser.parse_arguments(["--type", "Scan-Config"])
        self.assertEqual(args.type, "scan-config")
        args = parser.parse_arguments(["--type", "scan-configs"])
        self.assertEqual(args.type, "scan-config")
        args = parser.parse_arguments(["--type", "SCAN_CONFIGS"])
        self.assertEqual(args.type, "scan-config")

        args = parser.parse_arguments(["--type", "port-list"])
        self.assertEqual(args.type, "port-list")
        args = parser.parse_arguments(["--type", "PORT-LIST"])
        self.assertEqual(args.type, "port-list")
        args = parser.parse_arguments(["--type", "port_list"])
        self.assertEqual(args.type, "port-list")
        args = parser.parse_arguments(["--type", "PORT_LIST"])
        self.assertEqual(args.type, "port-list")
        args = parser.parse_arguments(["--type", "port-list"])
        self.assertEqual(args.type, "port-list")
        args = parser.parse_arguments(["--type", "port-lists"])
        self.assertEqual(args.type, "port-list")
        args = parser.parse_arguments(["--type", "PORT_LISTS"])
        self.assertEqual(args.type, "port-list")

        args = parser.parse_arguments(["--type", "gvmd-data"])
        self.assertEqual(args.type, "gvmd-data")
        args = parser.parse_arguments(["--type", "GVMD-DATA"])
        self.assertEqual(args.type, "gvmd-data")
        args = parser.parse_arguments(["--type", "gvmd_data"])
        self.assertEqual(args.type, "gvmd-data")
        args = parser.parse_arguments(["--type", "GVMD_DATA"])
        self.assertEqual(args.type, "gvmd-data")
        args = parser.parse_arguments(["--type", "gvmd-data"])
        self.assertEqual(args.type, "gvmd-data")

    def test_group(self):
        parser = CliParser()
        args = parser.parse_arguments(["--group", "some_group"])
        self.assertEqual(args.group, "some_group")

        args = parser.parse_arguments(["--group", "123"])
        self.assertEqual(args.group, 123)

    def test_user(self):
        parser = CliParser()
        args = parser.parse_arguments(["--user", "some_user"])
        self.assertEqual(args.user, "some_user")

        args = parser.parse_arguments(["--user", "123"])
        self.assertEqual(args.user, 123)

    def test_feed_url_from_enterprise_feed_key(self):
        parser = CliParser()
        content = """a_user@some.feed.server:/feed/
Lorem ipsum dolor sit amet,
consetetur sadipscing elitr,
sed diam nonumy eirmod tempor
"""
        with temp_file(content=content, name="enterprise.key") as f:
            args = parser.parse_arguments(
                ["--greenbone-enterprise-feed-key", str(f)]
            )

            self.assertEqual(
                args.feed_url, "ssh://a_user@some.feed.server/enterprise"
            )
            self.assertEqual(
                args.gvmd_data_url,
                f"ssh://a_user@some.feed.server/enterprise/data-feed/{DEFAULT_FEED_RELEASE}/",
            )
            self.assertEqual(
                args.port_lists_url,
                f"ssh://a_user@some.feed.server/enterprise/data-feed/{DEFAULT_FEED_RELEASE}/port-lists/",
            )
            self.assertEqual(
                args.report_formats_url,
                f"ssh://a_user@some.feed.server/enterprise/data-feed/{DEFAULT_FEED_RELEASE}/report-formats/",
            )
            self.assertEqual(
                args.scan_configs_url,
                f"ssh://a_user@some.feed.server/enterprise/data-feed/{DEFAULT_FEED_RELEASE}/scan-configs/",
            )
            self.assertEqual(
                args.notus_url,
                f"ssh://a_user@some.feed.server/enterprise/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
            )
            self.assertEqual(
                args.nasl_url,
                f"ssh://a_user@some.feed.server/enterprise/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
            )
            self.assertEqual(
                args.scap_data_url,
                f"ssh://a_user@some.feed.server/enterprise/vulnerability-feed/{DEFAULT_FEED_RELEASE}/scap-data/",
            )
            self.assertEqual(
                args.cert_data_url,
                f"ssh://a_user@some.feed.server/enterprise/vulnerability-feed/{DEFAULT_FEED_RELEASE}/cert-data/",
            )

    def test_ignore_non_existing_enterprise_feed_key(self):
        parser = CliParser()
        args = parser.parse_arguments(
            ["--greenbone-enterprise-feed-key", "/tmp/some.key"]
        )

        self.assertEqual(
            args.feed_url, "rsync://feed.community.greenbone.net/community"
        )
        self.assertEqual(
            args.gvmd_data_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{DEFAULT_FEED_RELEASE}/",
        )
        self.assertEqual(
            args.port_lists_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{DEFAULT_FEED_RELEASE}/port-lists/",
        )
        self.assertEqual(
            args.report_formats_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{DEFAULT_FEED_RELEASE}/report-formats/",
        )
        self.assertEqual(
            args.scan_configs_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{DEFAULT_FEED_RELEASE}/scan-configs/",
        )
        self.assertEqual(
            args.notus_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/notus/",
        )
        self.assertEqual(
            args.nasl_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{DEFAULT_FEED_RELEASE}/vt-data/nasl/",
        )
        self.assertEqual(
            args.scap_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{DEFAULT_FEED_RELEASE}/scap-data/",
        )
        self.assertEqual(
            args.cert_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{DEFAULT_FEED_RELEASE}/cert-data/",
        )

    @patch.object(sys, "argv", ["greenbone-nvt-sync"])
    def test_greenbone_nvt_sync(self):
        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.type, "nvt")

    @patch.object(sys, "argv", ["greenbone-scapdata-sync"])
    def test_greenbone_scap_data_sync(self):
        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.type, "scap")

    @patch.object(sys, "argv", ["greenbone-certdata-sync"])
    def test_greenbone_cert_data_sync(self):
        parser = CliParser()
        args = parser.parse_arguments([])

        self.assertEqual(args.type, "cert")

    def test_feed_release(self):
        parser = CliParser()
        feed_release = "1.2.3"
        args = parser.parse_arguments(["--feed-release", feed_release])
        self.assertTrue(args.feed_release, "1.2.3")

        self.assertEqual(
            args.feed_url, "rsync://feed.community.greenbone.net/community"
        )
        self.assertEqual(
            args.gvmd_data_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/",
        )
        self.assertEqual(
            args.port_lists_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/port-lists/",
        )
        self.assertEqual(
            args.report_formats_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/report-formats/",
        )
        self.assertEqual(
            args.scan_configs_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/scan-configs/",
        )
        self.assertEqual(
            args.notus_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/vt-data/notus/",
        )
        self.assertEqual(
            args.nasl_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/vt-data/nasl/",
        )
        self.assertEqual(
            args.scap_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/scap-data/",
        )
        self.assertEqual(
            args.cert_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/cert-data/",
        )

    def test_feed_version(self):
        parser = CliParser()
        feed_release = "1.2.3"
        args = parser.parse_arguments(["--feed-version", feed_release])
        self.assertTrue(args.feed_version, "1.2.3")
        self.assertTrue(args.feed_release, "1.2.3")

        self.assertEqual(
            args.feed_url, "rsync://feed.community.greenbone.net/community"
        )
        self.assertEqual(
            args.gvmd_data_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/",
        )
        self.assertEqual(
            args.port_lists_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/port-lists/",
        )
        self.assertEqual(
            args.report_formats_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/report-formats/",
        )
        self.assertEqual(
            args.scan_configs_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/scan-configs/",
        )
        self.assertEqual(
            args.notus_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/vt-data/notus/",
        )
        self.assertEqual(
            args.nasl_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/vt-data/nasl/",
        )
        self.assertEqual(
            args.scap_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/scap-data/",
        )
        self.assertEqual(
            args.cert_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/cert-data/",
        )

    def test_feed_release_overrules_version(self):
        parser = CliParser()
        feed_release = "1.2.3"
        args = parser.parse_arguments(
            ["--feed-version", "2.3.4", "--feed-release", feed_release]
        )
        self.assertTrue(args.feed_version, "2.3.4")
        self.assertTrue(args.feed_release, "1.2.3")

        self.assertEqual(
            args.feed_url, "rsync://feed.community.greenbone.net/community"
        )
        self.assertEqual(
            args.gvmd_data_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/",
        )
        self.assertEqual(
            args.port_lists_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/port-lists/",
        )
        self.assertEqual(
            args.report_formats_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/report-formats/",
        )
        self.assertEqual(
            args.scan_configs_url,
            f"rsync://feed.community.greenbone.net/community/data-feed/{feed_release}/scan-configs/",
        )
        self.assertEqual(
            args.notus_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/vt-data/notus/",
        )
        self.assertEqual(
            args.nasl_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/vt-data/nasl/",
        )
        self.assertEqual(
            args.scap_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/scap-data/",
        )
        self.assertEqual(
            args.cert_data_url,
            f"rsync://feed.community.greenbone.net/community/vulnerability-feed/{feed_release}/cert-data/",
        )

    def test_destination_prefix(self):
        parser = CliParser()
        destination_prefix = "/tmp/foo/bar"
        args = parser.parse_arguments(
            ["--destination-prefix", destination_prefix]
        )

        self.assertEqual(args.type, "all")
        self.assertEqual(args.destination_prefix, Path(destination_prefix))
        self.assertEqual(
            args.gvmd_data_destination,
            Path(destination_prefix) / "gvm" / "data-objects" / "gvmd",
        )
        self.assertEqual(
            args.notus_destination,
            Path(destination_prefix) / "notus",
        )
        self.assertEqual(
            args.nasl_destination,
            Path(destination_prefix) / "openvas" / "plugins",
        )
        self.assertEqual(
            args.scap_data_destination,
            Path(destination_prefix) / "gvm" / "scap-data",
        )
        self.assertEqual(
            args.cert_data_destination,
            Path(destination_prefix) / "gvm" / "cert-data",
        )
        self.assertEqual(
            args.report_formats_destination,
            Path(destination_prefix)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "report-formats",
        )
        self.assertEqual(
            args.scan_configs_destination,
            Path(destination_prefix)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "scan-configs",
        )
        self.assertEqual(
            args.port_lists_destination,
            Path(destination_prefix)
            / "gvm"
            / "data-objects"
            / "gvmd"
            / "port-lists",
        )
        self.assertEqual(
            args.gvmd_lock_file,
            Path(destination_prefix) / DEFAULT_GVMD_LOCK_FILE_PATH,
        )
        self.assertEqual(
            args.openvas_lock_file,
            Path(destination_prefix) / DEFAULT_OPENVAS_LOCK_FILE_PATH,
        )



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import Iterable, Optional


class GreenboneFeedSyncError(Exception):
    """
    Base exception class
    """


class ConfigError(GreenboneFeedSyncError):
    """
    Error while processing a config
    """


class ConfigFileError(ConfigError):
    """
    Error while processing a config file
    """


class ExecProcessError(GreenboneFeedSyncError):
    """
    Error while executing a process
    """

    def __init__(
        self,
        returncode: int,
        cmd: Iterable[str],
        *,
        stdout: Optional[bytes] = None,
        stderr: Optional[bytes] = None,
    ) -> None:
        self.returncode = returncode
        self.cmd = cmd
        self.stout = (
            None if not stdout else stdout.decode("utf8", errors="ignore")
        )
        self.stderr = (
            None if not stderr else stderr.decode("utf8", errors="ignore")
        )

    def __str__(self):
        cmd = " ".join(self.cmd)
        return f"'{cmd}' returned non-zero exit status {self.returncode}."


class RsyncError(ExecProcessError):
    """
    Error while running rsync
    """

    def __init__(
        self,
        returncode: int,
        args: Iterable[str],
        stderr: Optional[bytes] = None,
    ) -> None:
        super().__init__(returncode, cmd=["rsync"] + list(args), stderr=stderr)


class FileLockingError(GreenboneFeedSyncError):
    """
    An error during locking a file
    """



# SPDX-FileCopyrightText: 2022-2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import asyncio
import os
from pathlib import Path
from typing import Iterable, Optional, Union
from urllib.parse import urlsplit

from greenbone.feed.sync.errors import RsyncError


async def exec_rsync(*args: str) -> None:
    """
    Run rsync

    Argument:
        args: Arguments for rsync
    """
    process = await asyncio.create_subprocess_exec(
        "rsync", *args, stderr=asyncio.subprocess.PIPE
    )
    _, stderr = await process.communicate()
    returncode = await process.wait()
    if returncode:
        raise RsyncError(returncode, args, stderr=stderr)


DEFAULT_RSYNC_URL = "rsync://feed.community.greenbone.net/community"
DEFAULT_RSYNC_COMPRESSION_LEVEL = 9
DEFAULT_RSYNC_TIMEOUT: Optional[int] = (
    None  # in seconds. 0 means no timeout and None use rsync default
)
DEFAULT_RSYNC_SSH_PORT = 24
DEFAULT_RSYNC_SSH_OPTS = (
    "-o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no"
)

PathLike = Union[os.PathLike, str]


class Rsync:
    """
    Class to sync the feed data via rsync

    Args:
        verbose: Enable verbose output
        private_subdir: A private directory to exclude from from the sync
        compression_level: Set an compression level explicitly.
            Default is 9 (highest).
        timeout: Set a specific timeout in seconds. Default timeout of rsync is
            used of not set explicitly. 0 for no timeout.
        ssh_key: SSH key for using ssh as rsync transport protocol.
        exclude: An iterable of directories to exclude from the sync.

    """

    def __init__(
        self,
        *,
        verbose: bool = False,
        private_subdir: Optional[PathLike] = None,
        compression_level: Optional[int] = DEFAULT_RSYNC_COMPRESSION_LEVEL,
        timeout: Optional[int] = DEFAULT_RSYNC_TIMEOUT,
        ssh_key: Optional[PathLike] = None,
        exclude: Optional[Iterable[PathLike]] = None,
    ) -> None:
        self.verbose = verbose
        self.private_subdir = private_subdir
        self.compression_level = compression_level
        self.timeout = timeout
        self.ssh_key = ssh_key
        self.exclude = exclude

    async def sync(self, url: str, destination: PathLike) -> None:
        """
        Sync data from a remote URL to a destination path

        Args:
            url: URL to sync
            destination: Path to store the downloaded data
        """
        dest = Path(destination)
        dest.mkdir(parents=True, exist_ok=True)
        splitted_url = urlsplit(url)

        rsync_default_options = [
            "--links",
            "--times",
            "--omit-dir-times",
            "--recursive",
            "--partial",
            "--progress",
        ]

        if "ssh" in splitted_url.scheme:
            port = splitted_url.port or DEFAULT_RSYNC_SSH_PORT
            # we use ssh now
            rsync_ssh_options = [
                "-e",
                f"ssh {DEFAULT_RSYNC_SSH_OPTS} -p {port} -i '{self.ssh_key}'",
            ]
            url = f"{splitted_url.netloc}:{splitted_url.path}"
        else:
            rsync_ssh_options = []

        rsync_timeout = (
            [
                f"--timeout={self.timeout}",
            ]
            if self.timeout is not None
            else []
        )

        rsync_compress = (
            [
                f"--compress-level={self.compression_level}",
            ]
            if self.compression_level is not None
            else []
        )

        rsync_delete = [
            "--delete",
        ]

        rsync_chmod = [
            "--perms",
            "--chmod=Fugo+r,Fug+w,Dugo-s,Dugo+rx,Dug+w",
        ]

        rsync_links = [
            "--copy-unsafe-links",
            "--hard-links",
        ]

        if self.private_subdir:
            rsync_delete.extend(["--exclude", os.fspath(self.private_subdir)])

        if self.exclude:
            for exclude in self.exclude:
                rsync_delete.extend(["--exclude", os.fspath(exclude)])

        rsync_verbose = ["-v"] if self.verbose else ["-q"]

        args = (
            rsync_default_options
            + rsync_ssh_options
            + rsync_timeout
            + rsync_verbose
            + rsync_compress
            + rsync_delete
            + rsync_chmod
            + rsync_links
            + [url, str(dest.absolute())]
        )

        await exec_rsync(*args)
